from __future__ import annotations

import logging
from base64 import b64decode
from typing import TYPE_CHECKING

try:
    from Crypto.PublicKey import ECC, RSA

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False

from flow.record.fieldtypes import posix_path, windows_path

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.fsutil import TargetPath, open_decompress
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.ssh.ssh import (
    KnownHostRecord,
    SSHPlugin,
    calculate_fingerprints,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime
    from pathlib import Path

    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

log = logging.getLogger(__name__)

PuTTYUserRecordDescriptor = create_extended_descriptor([UserRecordDescriptorExtension])
PuTTYSessionRecord = PuTTYUserRecordDescriptor(
    "application/putty/saved_session",
    [
        ("datetime", "ts"),
        ("string", "session_name"),
        ("string", "protocol"),
        ("string", "host"),
        ("string", "user"),
        ("varint", "port"),
        ("string", "remote_command"),
        ("string", "port_forward"),
        ("string", "manual_ssh_host_keys"),
        ("path", "path"),
    ],
)


class PuTTYPlugin(SSHPlugin):
    """Extract artifacts from the PuTTY client.

    NOTE:
        - Does not parse ``$HOME/.putty/randomseed`` (GNU/Linux)
          and ``HKCU\\Software\\SimonTatham\\PuTTY\\RandSeedFile`` (Windows)

    References:
        - http://www.chiark.greenend.org.uk/~sgtatham/putty/0.78/puttydoc.txt
        - http://www.chiark.greenend.org.uk/~sgtatham/putty/faq.html#faq-settings
    """

    __namespace__ = "putty"

    def __init__(self, target: Target):
        super().__init__(target)

        self.regf_installs, self.path_installs = self._detect_putty()

    def _detect_putty(
        self,
    ) -> tuple[list[set[RegistryKey, UserDetails | None]], list[set[TargetPath, UserDetails | None]]]:
        regf_installs, path_installs = [], []

        if self.target.has_function("registry"):
            regf_installs = [
                (key, self.target.registry.get_user_details(key))
                for key in self.target.registry.keys("HKCU\\Software\\SimonTatham\\PuTTY")
            ]

        path_installs = [
            (putty_path, user_details)
            for user_details in self.target.user_details.all_with_home()
            if (putty_path := user_details.home_path.joinpath(".putty")).exists()
        ]

        return regf_installs, path_installs

    def check_compatible(self) -> None:
        if not any(self.regf_installs + self.path_installs):
            raise UnsupportedPluginError("No PuTTY installations found")

    @export(record=KnownHostRecord)
    def known_hosts(self) -> Iterator[KnownHostRecord]:
        """Parse PuTTY saved SshHostKeys."""

        for putty_key, user_details in self.regf_installs:
            yield from self._regf_known_hosts(putty_key, user_details)

        for putty_path, user_details in self.path_installs:
            yield from self._path_known_hosts(putty_path, user_details)

    def _regf_known_hosts(self, putty_key: RegistryKey, user_details: UserDetails) -> Iterator[KnownHostRecord]:
        """Parse PuTTY traces in Windows registry."""

        try:
            ssh_host_keys = putty_key.subkey("SshHostKeys")
        except RegistryKeyNotFoundError:
            return

        for entry in ssh_host_keys.values():
            key_type, host = entry.name.split("@")
            port, host = host.split(":")

            public_key, fingerprints = construct_public_key(key_type, entry.value)

            yield KnownHostRecord(
                mtime_ts=ssh_host_keys.ts,
                host=host,
                port=port,
                key_type=key_type,
                public_key=public_key,
                fingerprint=fingerprints,
                comment="",
                marker=None,
                path=windows_path(ssh_host_keys.path),
                _target=self.target,
                _user=user_details.user if user_details else None,
            )

    def _path_known_hosts(self, putty_path: TargetPath, user_details: UserDetails) -> Iterator[KnownHostRecord]:
        """Parse PuTTY traces in ``.putty`` folders."""
        ssh_host_keys_path = putty_path.joinpath("sshhostkeys")

        if ssh_host_keys_path.exists():
            ts = ssh_host_keys_path.stat().st_mtime

            for line in open_decompress(ssh_host_keys_path, "rt"):
                parts = line.split()
                key_type, host = parts[0].split("@")
                port, host = host.split(":")

                public_key, fingerprints = construct_public_key(key_type, parts[1])

                yield KnownHostRecord(
                    mtime_ts=ts,
                    host=host,
                    port=port,
                    key_type=key_type,
                    public_key=public_key,
                    fingerprint=fingerprints,
                    comment="",
                    marker=None,
                    path=posix_path(ssh_host_keys_path),
                    _target=self.target,
                    _user=user_details.user if user_details else None,
                )

    @export(record=PuTTYSessionRecord)
    def sessions(self) -> Iterator[PuTTYSessionRecord]:
        """Parse PuTTY saved session configuration files."""

        for putty_key, user_details in self.regf_installs:
            yield from self._regf_sessions(putty_key, user_details)

        for putty_path, user_details in self.path_installs:
            yield from self._path_sessions(putty_path, user_details)

    def _regf_sessions(self, putty_key: RegistryKey, user_details: UserDetails) -> Iterator[PuTTYSessionRecord]:
        try:
            sessions = putty_key.subkey("Sessions")
        except RegistryKeyNotFoundError:
            return

        for session in sessions.subkeys():
            cfg = {s.name: s.value for s in session.values()}
            yield from self._build_session_record(
                session.ts, session.name, windows_path(session.path), cfg, user_details
            )

    def _path_sessions(self, putty_path: TargetPath, user_details: UserDetails) -> Iterator[PuTTYSessionRecord]:
        sessions_dir = putty_path.joinpath("sessions")
        if sessions_dir.exists():
            for session in sessions_dir.glob("*"):
                if session.is_file():
                    cfg = dict(map(str.strip, line.split("=", maxsplit=1)) for line in session.open("rt").readlines())
                    yield from self._build_session_record(
                        session.stat().st_mtime, session.name, session, cfg, user_details
                    )

    def _build_session_record(
        self, ts: float, name: float | datetime, source: Path, cfg: dict, user_details: UserDetails
    ) -> PuTTYSessionRecord:
        host, user = parse_host_user(cfg.get("HostName"), cfg.get("UserName"))

        yield PuTTYSessionRecord(
            ts=ts,
            session_name=name,
            protocol=cfg.get("Protocol"),
            host=host,
            user=user,
            port=cfg.get("PortNumber"),
            remote_command=cfg.get("RemoteCommand"),
            port_forward=cfg.get("PortForwardings"),
            manual_ssh_host_keys=cfg.get("SSHManualHostKeys"),
            path=source,
            _target=self.target,
            _user=user_details.user if user_details else None,
        )


def parse_host_user(host: str, user: str) -> tuple[str, str]:
    """Parse host and user from PuTTY hostname component."""
    if "@" in host:
        parsed_user, parsed_host = host.split("@")
        user = user or parsed_user
        host = parsed_host

    return host, user


def construct_public_key(key_type: str, iv: str) -> tuple[str, tuple[str, str, str]]:
    """Returns OpenSSH format public key calculated from PuTTY SshHostKeys format and set of fingerprints.

    PuTTY stores raw public key components instead of OpenSSH-formatted public keys
    or fingerprints. With RSA public keys the exponent and modulus are stored.
    With ECC keys the x and y prime coordinates are stored together with the curve type.

    Currently supports ``ssh-ed25519``, ``ecdsa-sha2-nistp256`` and ``rsa2`` key types.

    NOTE: More key types could be supported in the future.

    References:
        - https://github.com/github/putty/blob/master/contrib/kh2reg.py
        - https://pycryptodome.readthedocs.io/en/latest/src/public_key/rsa.html
        - https://pycryptodome.readthedocs.io/en/latest/src/public_key/ecc.html
        - https://github.com/mkorthof/reg2kh
    """
    if not HAS_CRYPTO:
        log.warning("Could not reconstruct public key: missing pycryptodome dependency")
        return iv, (None, None, None)

    if not isinstance(key_type, str) or not isinstance(iv, str):
        raise TypeError("Invalid key_type or iv")

    key = None

    if key_type == "ssh-ed25519":
        x, y = iv.split(",")
        key = ECC.construct(curve="ed25519", point_x=int(x, 16), point_y=int(y, 16))

    if key_type == "ecdsa-sha2-nistp256":
        _, x, y = iv.split(",")
        key = ECC.construct(curve="NIST P-256", point_x=int(x, 16), point_y=int(y, 16))

    if key_type == "rsa2":
        exponent, modulus = iv.split(",")
        key = RSA.construct((int(modulus, 16), int(exponent, 16)))

    if key is None:
        log.warning("Could not reconstruct public key: type %s not implemented", key_type)
        return iv, (None, None, None)

    openssh_public_key = key.public_key().export_key(format="OpenSSH")

    if isinstance(openssh_public_key, bytes):
        # RSA's export_key() returns bytes
        openssh_public_key = openssh_public_key.decode()

    key_part = openssh_public_key.split()[-1]
    fingerprints = calculate_fingerprints(b64decode(key_part))
    return key_part, fingerprints



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.applications import UnixApplicationsPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_unix_applications_desktop_files(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if .desktop files registering installed applications are detected correctly."""

    fs_unix.map_file_fh("/etc/hostname", BytesIO(b"hostname"))

    # system paths
    fs_unix.map_file(
        "/var/lib/snapd/desktop/applications/firefox_firefox.desktop",
        absolute_path("_data/plugins/os/unix/applications/firefox_firefox.desktop"),
    )
    fs_unix.map_file(
        "/var/lib/snapd/desktop/applications/code_code.desktop",
        absolute_path("_data/plugins/os/unix/applications/code_code.desktop"),
    )
    fs_unix.map_file(
        "/usr/share/applications/gimp.desktop",
        absolute_path("_data/plugins/os/unix/applications/gimp.desktop"),
    )
    fs_unix.map_file(
        "/usr/local/share/applications/vmware-workstation.desktop",
        absolute_path("_data/plugins/os/unix/applications/vmware-workstation.desktop"),
    )
    fs_unix.map_file(
        "/var/lib/flatpak/exports/share/applications/python.desktop",
        absolute_path("_data/plugins/os/unix/applications/python.desktop"),
    )

    # user paths
    fs_unix.map_file(
        "/home/user/.local/share/applications/vlc.desktop",
        absolute_path("_data/plugins/os/unix/applications/vlc.desktop"),
    )
    fs_unix.map_file(
        "/root/.local/share/applications/terminal.desktop",
        absolute_path("_data/plugins/os/unix/applications/terminal.desktop"),
    )

    target_unix_users.add_plugin(UnixPlugin)
    target_unix_users.add_plugin(UnixApplicationsPlugin)
    results = sorted(target_unix_users.applications(), key=lambda r: r.name)

    assert len(results) == 7

    assert results[0].ts_installed is None
    assert results[0].name == "Firefox Web Browser"
    assert results[0].version == "1.0"
    assert results[0].author is None
    assert results[0].type == "user"
    assert (
        results[0].path
        == "env BAMF_DESKTOP_FILE_HINT=/var/lib/snapd/desktop/applications/firefox_firefox.desktop /snap/bin/firefox %u"
    )
    assert results[0].hostname == "hostname"

    assert [r.name for r in results] == [
        "Firefox Web Browser",
        "GNU Image Manipulation Program",
        "Python (v3.12)",
        "Terminal",
        "VLC media player",
        "VMware Workstation",
        "Visual Studio Code",
    ]

    assert [r.path for r in results] == [
        "env BAMF_DESKTOP_FILE_HINT=/var/lib/snapd/desktop/applications/firefox_firefox.desktop /snap/bin/firefox %u",
        "gimp-2.10 %U",
        "/usr/bin/python3.12",
        "gnome-terminal",
        "/usr/bin/vlc --started-from-file %U",
        "/usr/bin/vmware %U",
        "env BAMF_DESKTOP_FILE_HINT=/var/lib/snapd/desktop/applications/code_code.desktop /snap/bin/code --force-user-env %F",  # noqa: E501
    ]

    assert [r.type for r in results] == [
        "user",
        "user",
        "user",
        "system",
        "user",
        "user",
        "user",
    ]



from __future__ import annotations

import bisect
import io
from pathlib import Path
from typing import BinaryIO

from dissect.target.container import Container


def find_files(path: Path) -> list[Path]:
    path = path.resolve()
    return sorted([f for f in path.parent.glob(path.stem + ".*") if f.suffix[1:].isdigit()])


class SplitContainer(Container):
    __type__ = "split"

    def __init__(self, fh: list | BinaryIO | Path, *args, **kwargs):
        self._fhs = []
        self.offsets = [0]
        offset = 0

        fhs = [fh] if not isinstance(fh, list) else fh
        if isinstance(fhs[0], Path):
            fhs = [path.open("rb") for path in find_files(fhs[0])]

        for f in fhs:
            f.seek(0, io.SEEK_END)

            offset += f.tell()
            self.offsets.append(offset)
            self._fhs.append(f)
        self._roffset = 0

        super().__init__(fh, offset, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return isinstance(original, list) and len(original) > 1

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return (path.suffix[1:].isdigit() and len(find_files(path)) > 1) or (
            isinstance(original, list) and len(original) > 1
        )

    def read(self, length: int) -> bytes:
        length = min(length, self.size - self._roffset)
        buf = b""

        while length > 0:
            data = self._read_partial(length)
            if not data:
                break

            length -= len(data)
            buf += data
            self._roffset += len(data)

        return buf

    def _read_partial(self, length: int) -> bytes:
        idx = bisect.bisect_right(self.offsets, self._roffset + 1) - 1
        fh = self._fhs[idx]

        img_offset = self.offsets[idx]
        fh.seek(self._roffset - img_offset)

        return fh.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if whence is io.SEEK_SET:
            self._roffset = offset
        elif whence is io.SEEK_CUR:
            self._roffset += offset
        elif whence is io.SEEK_END:
            self._roffset = self.size + offset

        self._roffset = min(self._roffset, self.size)
        return self._roffset

    def tell(self) -> int:
        return self._roffset

    def close(self) -> None:
        pass



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows.credential.lsa import LSAPlugin
from tests.plugins.os.windows.test__os import map_version_value

if TYPE_CHECKING:
    from dissect.target.target import Target

SYSTEM_KEY = "SYSTEM\\ControlSet001\\Control\\LSA"
POLICY_KEY_PATH_NT5 = "SECURITY\\Policy\\PolSecretEncryptionKey"
POLICY_KEY_PATH_NT6 = "SECURITY\\Policy\\PolEKList"
SECRETS_KEY = "SECURITY\\Policy\\Secrets"


def map_lsa_system_keys(hive_hklm: VirtualHive, subkeys: dict) -> None:
    """Add values to the registry required to calculate the SYSKEY / BootKey."""

    if subkeys.keys() != {"Data", "GBG", "JD", "Skew1"}:
        raise ValueError("Invalid subkey names")

    system_key = VirtualKey(hive_hklm, SYSTEM_KEY)

    for name, value in subkeys.items():
        system_key.add_subkey(name, VirtualKey(hive_hklm, name, class_name=value))

    hive_hklm.map_key(SYSTEM_KEY, system_key)


def map_lsa_polkey(hive_hklm: VirtualHive, path: str, value: bytes) -> None:
    """Add policy key to the registry which is required to derive the LSA key of the system."""

    policy_key = VirtualKey(hive_hklm, path)
    policy_key.add_value("(Default)", value)
    hive_hklm.map_key(path, policy_key)


def map_lsa_secrets(hive_hklm: VirtualHive, secrets: dict[str, bytes | tuple[bytes, bytes]]) -> None:
    """Add given encrypted LSA secrets to the ``hive_hklm`` :class"`VirtualHive`."""

    secrets_key = VirtualKey(hive_hklm, SECRETS_KEY)

    for name, value in secrets.items():
        if not isinstance(value, (bytes, tuple)):
            raise TypeError(f"Given value for {name} should be in bytes!")

        if isinstance(value, tuple):
            curr_val, old_val = value
        else:
            curr_val = value
            old_val = None

        secret_key = VirtualKey(hive_hklm, name)
        secret_key.timestamp = datetime(2025, 12, 31, 23, 59, 59, tzinfo=timezone.utc)

        currval_key = VirtualKey(hive_hklm, "CurrVal")
        currval_key.add_value("(Default)", curr_val)
        secret_key.add_subkey("CurrVal", currval_key)

        if old_val:
            oldval_key = VirtualKey(hive_hklm, "OldVal")
            oldval_key.add_value("(Default)", old_val)
            secret_key.add_subkey("OldVal", oldval_key)

        secrets_key.add_subkey(name, secret_key)

    hive_hklm.map_key(SECRETS_KEY, secrets_key)


def test_lsa_secrets_win_10(target_win: Target, hive_hklm: VirtualHive) -> None:
    """Test decrypting LSA secrets of a Windows 10 system."""

    map_lsa_system_keys(
        hive_hklm,
        {
            "Data": "3b03ce4f",
            "GBG": "7bf5e093",
            "JD": "51fe8c74",
            "Skew1": "3e14f655",
        },
    )

    policy_key = VirtualKey(hive_hklm, POLICY_KEY_PATH_NT6)
    policy_key_value = (
        "00000001ecffe17b2a997440aa939adbff26f1fc030000000000000061f61040"
        "72b7b9117289a891bc1edb9fa83fafae5d8648b8674825dc9bdcf1bf5ffbe591"
        "27fd3a1baca6c5447530574c27a41ab721b13e54a9af857e682b8458859c0fab"
        "55461a8f1b2b8c12d8f186a80991012e2d3ede5b1167554c726e404ed0eb96f6"
        "488b90ece5b0b3ef9bf244f35dd65a30b2b450c4e227838810b631564b89300e"
        "d4355040f23890f8b6b8eab8"
    )
    policy_key.add_value("(Default)", bytes.fromhex(policy_key_value))
    hive_hklm.map_key(POLICY_KEY_PATH_NT6, policy_key)

    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": (
                # CurrVal
                bytes.fromhex(
                    "00000001001f9b85984f68a8ed3e9d44dbd5b79c0300000000000000835b4ff6"
                    "6e74154ffab75afd31a2860616c87411bc97d368a068fca62d1564ae1396f88e"
                    "06de87b5e5632c668538ee36c75f67cee98d49ebc1e88fa7e9be16144af31e8e"
                    "5fd78329279e50d792e8a6b35a59cb55016748ecd8e12f148b1d32b3"
                ),
                # OldVal
                bytes.fromhex(
                    "00000001001f9b85984f68a8ed3e9d44dbd5b79c0300000000000000835b4ff6"
                    "6e74154ffab75afd31a2860616c87411bc97d368a068fca62d1564ae1396f88e"
                    "06de87b5e5632c668538ee36c75f67cee98d49ebc1e88fa7e9be16144af31e8e"
                    "5fd78329279e50d792e8a6b35a59cb55016748ecd8e12f148b1d32b3"
                ),
            ),
        },
    )

    map_version_value(target_win, "CurrentVersion", 10.0)
    target_win.add_plugin(LSAPlugin)

    assert target_win.lsa.syskey.hex() == "7b143e8c93f5037451f6fe3bcee04f55"
    assert target_win.lsa.lsakey.hex() == "26c34326a9f1d8db4f465487857d47d1481f0040c6becc8c65862c0cfb210631"
    assert target_win.lsa._secrets == {
        "DPAPI_SYSTEM": bytes.fromhex(
            "2c000000000000000000000000000000010000003ac5746c27b424489300a781"
            "ba237676da7605083e7947c67271c16fc84cc567870148931cf29eb800000000"
        ),
        "DPAPI_SYSTEM_OldVal": bytes.fromhex(
            "2c000000000000000000000000000000010000003ac5746c27b424489300a781"
            "ba237676da7605083e7947c67271c16fc84cc567870148931cf29eb800000000"
        ),
    }

    records = list(target_win.lsa.secrets())
    assert len(records) == 2

    assert records[0].ts == datetime(2025, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    assert records[0].name == "DPAPI_SYSTEM"
    assert (
        records[0].value
        == "2c000000000000000000000000000000010000003ac5746c27b424489300a781ba237676da7605083e7947c67271c16fc84cc567870148931cf29eb800000000"  # noqa: E501
    )

    assert records[1].ts == datetime(2025, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    assert records[1].name == "DPAPI_SYSTEM_OldVal"
    assert (
        records[1].value
        == "2c000000000000000000000000000000010000003ac5746c27b424489300a781ba237676da7605083e7947c67271c16fc84cc567870148931cf29eb800000000"  # noqa: E501
    )


def test_lsa_secrets_win_xp(target_win: Target, hive_hklm: VirtualHive) -> None:
    """Test decrypting LSA secrets of a Windows XP system."""

    map_lsa_system_keys(
        hive_hklm,
        {
            "Data": "0e1eaddb",
            "GBG": "36ec555b",
            "JD": "8397b619",
            "Skew1": "9dfad6c8",
        },
    )

    policy_key = VirtualKey(hive_hklm, POLICY_KEY_PATH_NT5)
    policy_key_value = (
        "0100000001000000000000003e4253c6ab72f73306ffe45a33a4ec5d47d21483"
        "85be6036b4c5bab2666d7f5793f23fed8513577b79abddb40eaa92c7a19b34ab"
        "b9c134e88bf5dd2b2fbf6356"
    )
    policy_key.add_value("(Default)", bytes.fromhex(policy_key_value))
    hive_hklm.map_key(POLICY_KEY_PATH_NT5, policy_key)

    map_lsa_secrets(
        hive_hklm,
        {
            "DPAPI_SYSTEM": bytes.fromhex(
                "3800000038000020acb90c00a26440bdf0382bdbcb5d51a743cb4cca01cfecb3"
                "202848e00dfae25f8f0f3f4710ff37828532b5f600e97f6a00528fd8e0137f49"
                "6fd2d905"
            ),
        },
    )

    map_version_value(target_win, "CurrentVersion", 5.1)
    target_win.add_plugin(LSAPlugin)

    assert target_win.lsa.syskey.hex() == "36fa9db65bec1e1983d6970ead55dbc8"
    assert target_win.lsa.lsakey.hex() == "e32851688cf1e05b1da350682a2e07d2"
    assert target_win.lsa._secrets == {
        "DPAPI_SYSTEM": bytes.fromhex(
            "2c00000001000000010000000307e7679d5bc2d7a2212e216d527af1687f7183"
            "333dac1af5cc9d33ccd97238f3a25ebfaec94ac600000000"
        )
    }

    records = list(target_win.lsa.secrets())
    assert len(records) == 1
    assert records[0].ts == datetime(2025, 12, 31, 23, 59, 59, tzinfo=timezone.utc)
    assert records[0].name == "DPAPI_SYSTEM"
    assert (
        records[0].value
        == "2c00000001000000010000000307e7679d5bc2d7a2212e216d527af1687f7183333dac1af5cc9d33ccd97238f3a25ebfaec94ac600000000"  # noqa: E501
    )



from __future__ import annotations

import gzip
import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.unix.log.helpers import is_iso_fmt, iso_readlines

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem

syslog = """\
Dec 31 03:14:15 localhost systemd[1]: Starting Journal Service...
Jan  1 13:21:34 localhost systemd: Stopped target Swap.
Jan  2 03:14:15 localhost systemd[1]: Starting Journal Service...
Jan  3 13:21:34 localhost systemd: Stopped target Swap.
2024-12-31T13:37:00.123456+02:00 hostname systemd[1]: Started anacron.service - Run anacron jobs.
2024-12-31T13:37:00.123456+02:00 hostname anacron[1337]: Anacron 2.3 started on 2024-12-31
2024-12-31T13:37:00.123456+02:00 hostname anacron[1337]: Normal exit (0 jobs run)
2024-12-31T13:37:00.123456+02:00 hostname systemd[1]: anacron.service: Deactivated successfully.
"""


@pytest.mark.parametrize(
    ("max_lines", "expected_return_value"),
    [
        (3, False),
        (4, False),
        (5, True),
        (9, True),
    ],
)
def test_iso_readlines_max_lines(fs_unix: VirtualFilesystem, max_lines: int, expected_return_value: bool) -> None:
    """Assert that iso_readlines does not parse more than the provided max_lines."""

    fs_unix.map_file_fh("/var/log/syslog.2", BytesIO(gzip.compress(textwrap.dedent(syslog).encode())))
    assert any(iso_readlines(fs_unix.path("/var/log/syslog.2"), max_lines)) == expected_return_value


def test_is_iso_fmt(fs_unix: VirtualFilesystem) -> None:
    """Assert that is_iso_fmt does not parse more than three max_lines."""

    fs_unix.map_file_fh("/var/log/syslog.3", BytesIO(gzip.compress(textwrap.dedent(syslog).encode())))
    assert not is_iso_fmt(fs_unix.path("/var/log/syslog.3"))



from __future__ import annotations

from pathlib import Path

from dissect.target.filesystems.vmtar import VmtarFilesystem
from tests._utils import absolute_path


def test_filesystems_vmtar() -> None:
    vmtar_path = Path(absolute_path("_data/filesystems/vmtar/simple.vmtar"))

    with vmtar_path.open("rb") as fh:
        assert VmtarFilesystem.detect(fh)

        fs = VmtarFilesystem(fh)
        assert [f.name for f in fs.path("/").iterdir()] == ["hello_world.txt"]
        assert fs.get("hello_world.txt").open().read() == b"hello_from_a_tar_file\n"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.bsd.darwin.ios._os import IOSPlugin
from dissect.target.plugins.os.unix.bsd.darwin.macos._os import MacOSPlugin
from dissect.target.plugins.os.unix.linux.network_managers import (
    LinuxNetworkManager,
    parse_unix_dhcp_leases,
    parse_unix_dhcp_log_messages,
)
from dissect.target.plugins.os.windows._os import WindowsPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class LinuxPlugin(UnixPlugin, LinuxNetworkManager):
    """Linux plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.network_manager = LinuxNetworkManager(target)
        self.network_manager.discover()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        """Detect a Linux-like filesystem.

        These days there is little difference in the filesystem format used by Unix and Linux. Both implementations use
        the Filesystem Hierarchy Standard (FHS). We can differentiate between Unix and Linux by checking for specific
        Linux kernel-only files not present on actual Unix filesystems (e.g. BSD, Solaris, IBM AIX and HP-UX).

        References:
            - https://refspecs.linuxfoundation.org/fhs.shtml
            - https://en.wikipedia.org/wiki/Filesystem_Hierarchy_Standard
        """

        # NOTE: dirs like /opt, /mnt, /media, /tmp and /proc are not Linux-specific.
        LINUX_PATHS = {
            "/run",
            "/sys",
            "/etc/kernel",
            "/etc/sysctl.conf",
            "/var/log/kern.log",
            "/boot/initrd.img",
            "/boot/vmlinuz",
            "/boot/vmlinux",
            "/opt",  # Backwards compatibility for previous Linux detection.
        }

        for fs in target.filesystems:
            # We explicitly exclude filesystems that look more like a macOS or Windows sysvol.
            if MacOSPlugin.detect(target) or IOSPlugin.detect(target) or WindowsPlugin.detect(target):
                continue

            # Dirs /var and /etc make this a Unix-like system (see UnixPlugin.detect),
            # while Linux-kernel specific files make it a Linux filesystem.
            if fs.exists("/var") and fs.exists("/etc") and any(fs.exists(p) for p in LINUX_PATHS):
                return fs

            # This filesystem could be a volatile collection of a Linux system.
            if fs.exists("/sys/module") or fs.exists("/proc/sys"):
                return fs
        return None

    @export(property=True)
    def ips(self) -> list[str]:
        """Returns a list of static IP addresses and DHCP lease IP addresses found on the host system."""
        ips = set()

        for ip_set in self.network_manager.get_config_value("ips"):
            ips.update(ip_set)

        if dhcp_lease_ips := parse_unix_dhcp_leases(self.target):
            ips.update(dhcp_lease_ips)

        elif dhcp_log_ips := parse_unix_dhcp_log_messages(self.target, iter_all=False):
            ips.update(dhcp_log_ips)

        return list(ips)

    @export(property=True)
    def dns(self) -> list[str]:
        return self.network_manager.get_config_value("dns")

    @export(property=True)
    def dhcp(self) -> list[bool]:
        return self.network_manager.get_config_value("dhcp")

    @export(property=True)
    def gateway(self) -> list[str]:
        return self.network_manager.get_config_value("gateway")

    @export(property=True)
    def interface(self) -> list[str]:
        return self.network_manager.get_config_value("interface")

    @export(property=True)
    def netmask(self) -> list[str]:
        return self.network_manager.get_config_value("netmask")

    @export(property=True)
    def version(self) -> str | None:
        distrib_description = self._os_release.get("DISTRIB_DESCRIPTION", "")
        name = self._os_release.get("NAME", "") or self._os_release.get("DISTRIB_ID", "")
        version = (
            self._os_release.get("VERSION", "")
            or self._os_release.get("VERSION_ID", "")
            or self._os_release.get("DISTRIB_RELEASE", "")
        )

        if not any([name, version, distrib_description]):
            return None

        if len(f"{name} {version}") > len(distrib_description):
            distrib_description = f"{name} {version}"

        return distrib_description or None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.LINUX.value



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.remoteaccess.rustdesk import RustdeskPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_rustdesk_plugin_log(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "Windows/ServiceProfiles/LocalService/AppData/Roaming/RustDesk/log/server/TestRustdesk.log",
        absolute_path("_data/plugins/apps/remoteaccess/rustdesk/TestRustdesk.log"),
    )

    target_win_users.add_plugin(RustdeskPlugin)

    records = list(target_win_users.rustdesk.logs())
    assert len(records) == 1

    assert records[0].ts == datetime(2025, 1, 1, 13, 4, 8, 350802, tzinfo=timezone.utc)
    assert records[0].message == "DEBUG src\\server\\connection.rs:983 #1362 Connection opened from REDACTED IP:6074."
    assert (
        records[0].source
        == "c:\\Windows\\ServiceProfiles\\LocalService\\AppData\\Roaming\\RustDesk\\log\\server\\TestRustdesk.log"
    )
    assert records[0].username is None
    assert records[0].user_id is None
    assert records[0].user_home is None


def test_rustdesk_plugin_user_log(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    fs_win.map_file(
        "Users/John/AppData/Roaming/Rustdesk/log/TestRustdesk.log",
        absolute_path("_data/plugins/apps/remoteaccess/rustdesk/TestRustdesk.log"),
    )

    target_win_users.add_plugin(RustdeskPlugin)

    records = list(target_win_users.rustdesk.logs())
    assert len(records) == 1
    assert records[0].ts == datetime(2025, 1, 1, 13, 4, 8, 350802, tzinfo=timezone.utc)
    assert records[0].message == "DEBUG src\\server\\connection.rs:983 #1362 Connection opened from REDACTED IP:6074."
    assert records[0].source == "C:/Users/John/AppData/Roaming/Rustdesk/log/TestRustdesk.log"



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.av.trendmicro import (
    TrendMicroPlugin,
    TrendMicroWFFirewallRecord,
    TrendMicroWFLogRecord,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_trendmicro_plugin_worryfree_firewall(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/trendmicro/firewall.log")
    fs_win.map_file("Program Files (x86)/Trend Micro/Security Agent/PFW/PfwLog_20230101.dat", log_file)
    target_win.add_plugin(TrendMicroPlugin)
    records = list(target_win.trendmicro.wffirewall())
    assert len(records) == 1
    assert isinstance(records[0], type(TrendMicroWFFirewallRecord()))
    assert records[0].ts == dt("2023-03-14T09:42:55Z")
    assert str(records[0].local_ip) == "127.0.0.1"
    assert str(records[0].remote_ip) == "255.255.255.255"
    assert records[0].direction == "in"
    assert records[0].port == 444
    assert str(records[0].path) == "C:\\WINDOWS\\SYSTEM32\\SVCHOST.EXE"
    assert records[0].description == "SecurityLevelDrop"


def test_trendmicro_plugin_worryfree_log(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/trendmicro/pccnt35.log")
    fs_win.map_file("Program Files (x86)/Trend Micro/Security Agent/Misc/pccnt35.log", log_file)
    target_win.add_plugin(TrendMicroPlugin)
    records = list(target_win.trendmicro.wflogs())
    assert len(records) == 1
    assert isinstance(records[0], type(TrendMicroWFLogRecord()))
    assert records[0].ts == dt("2023-03-10T15:06:19Z")
    assert records[0].threat == "Eicar_test_file"
    assert str(records[0].path) == "C:\\Users\\admin\\Desktop\\test\\eicarcom2.zip"
    assert records[0].lineno == 0



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.child.virtualbox import VirtualBoxChildTargetPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_child_virtualbox_linux(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we detect Oracle VirtualBox children on a Linux target."""

    fs_unix.map_file(
        "/home/user/.config/VirtualBox/VirtualBox.xml",
        absolute_path("_data/plugins/child/virtualbox/VirtualBox.xml"),
    )

    # vbox to be found by traversing MachineRegistry values
    fs_unix.map_file_fh("/example/vms/example-vm/example-vm.vbox", BytesIO())
    fs_unix.map_file_fh("/example/vms/second-vm/second-vm.vbox", BytesIO())
    fs_unix.map_file_fh("/example/vms/third-vm/third-vm.vbox", BytesIO())

    # vbox to be found by traversing SystemProperties defaultMachineFolder value
    fs_unix.map_file_fh("/some/other/folder/VirtualBox VMs/fourth-vm/fourth-vm.vbox", BytesIO())

    # vbox to be found by traversing `$HOME/VirtualBox VMs` folders
    fs_unix.map_file_fh("/home/user/VirtualBox VMs/fifth-vm/fifth-vm.vbox", BytesIO())

    # test deduplication by mapping the same VirtualBox.xml file for the root user
    fs_unix.map_file(
        "/root/.config/VirtualBox/VirtualBox.xml-prev",
        absolute_path("_data/plugins/child/virtualbox/VirtualBox.xml"),
    )

    target_unix_users.add_plugin(VirtualBoxChildTargetPlugin)
    children = list(target_unix_users.list_children())

    assert sorted(map(str, [child.path for child in children])) == [
        "/example/vms/example-vm/example-vm.vbox",
        "/example/vms/second-vm/second-vm.vbox",
        "/example/vms/third-vm/third-vm.vbox",
        "/home/user/VirtualBox VMs/fifth-vm/fifth-vm.vbox",
        "/some/other/folder/VirtualBox VMs/fourth-vm/fourth-vm.vbox",
    ]



from __future__ import annotations

import json
import re
from enum import Enum
from typing import TYPE_CHECKING

from dissect.sql import Error as SQLError
from dissect.sql import SQLite3

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.container.container import (
    COMMON_CONTAINER_FIELDS,
    COMMON_IMAGE_FIELDS,
    COMMON_LOG_FIELDS,
    ContainerPlugin,
)
from dissect.target.plugins.apps.container.docker import hash_to_image_id

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

PodmanImageRecord = TargetRecordDescriptor(
    "apps/containers/podman/image",
    COMMON_IMAGE_FIELDS,
)

PodmanContainerRecord = TargetRecordDescriptor(
    "apps/containers/podman/container",
    COMMON_CONTAINER_FIELDS,
)

PodmanLogRecord = TargetRecordDescriptor(
    "apps/containers/podman/log",
    COMMON_LOG_FIELDS,
)


class ContainerState(Enum):
    """Enum of possible Container states.

    References:
        - https://github.com/containers/podman/blob/v4.9/libpod/define/containerstate.go
    """

    NONE = None
    UNKNOWN = 0
    CONFIGURED = 1
    CREATED = 2
    RUNNING = 3
    STOPPED = 4
    PAUSED = 5
    EXITED = 6
    REMOVING = 7
    STOPPING = 8


RE_CTR_LOG = re.compile(
    r"""
        ^
        (?P<ts>\d{4}-\d{2}-\d{2}T\d{2}\:\d{2}\:\d{2}\.\d{9}\+\d{2}\:\d{2})
        \s
        (?P<stream>(stderr|stdout))
        \s
        (?P<type>\S)
        \s
        (?P<message>.*)
        $
    """,
    re.VERBOSE,
)


class PodmanPlugin(ContainerPlugin):
    """Parse Podman artefacts.

    References:
        - https://docs.podman.io/en/latest/_static/api.html
    """

    __namespace__ = "podman"

    SYSTEM_PATHS = (
        # Linux
        "/var/lib/containers",
        # Windows
        "sysvol/ProgramData/containers",
        "sysvol/Program Files/RedHat/Podman/containers",
    )

    USER_PATHS = (
        # Linux and Windows
        ".local/share/containers",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.installs = set(self.find_installs())

    def find_installs(self) -> Iterator[tuple[Path, UserDetails | None]]:
        for path in self.SYSTEM_PATHS:
            if (dir := self.target.fs.path(path)).exists():
                yield dir, None

        for user_details in self.target.user_details.all_with_home():
            for path in self.USER_PATHS:
                if (dir := user_details.home_path.joinpath(path)).exists():
                    yield dir, user_details

    def check_compatible(self) -> None:
        if not self.installs:
            raise UnsupportedPluginError("No Podman install folders found on target")

    @export(record=PodmanImageRecord)
    def images(self) -> Iterator[PodmanImageRecord]:
        """Yield any pulled Podman images on the target system."""
        for dir, _ in self.installs:
            images_json = dir.joinpath("storage/overlay-images/images.json")

            if not images_json.exists():
                self.target.log.debug("No Podman images found, file not found: %s", images_json)
                continue

            try:
                images = json.loads(images_json.read_text()) or []
            except (json.JSONDecodeError, UnicodeDecodeError) as e:
                self.target.log.warning("Unable to parse JSON in: %s", images_json)
                self.target.log.debug("", exc_info=e)
                continue

            for image in images:
                name = None
                tag = None

                if image_name := image.get("names", [None])[0]:
                    name, _, tag = image_name.rpartition(":")

                # TODO: when was the image pulled? It is not `created`.
                yield PodmanImageRecord(
                    name=name,
                    tag=tag,
                    image_id=hash_to_image_id(image.get("id")),
                    created=image.get("created"),
                    hash=image.get("id"),
                    source=images_json,
                    _target=self.target,
                )

    @export(record=PodmanContainerRecord)
    def containers(self) -> Iterator[PodmanContainerRecord]:
        """Yield any Podman containers on the target system.

        Uses the ``$PODMAN/storage/db.sql`` SQLite3 database and ``$PODMAN/storage/overlay-containers`` folder.
        Does not support legacy (v3 and older) BoltDB Podman databases.
        """
        for dir, _ in self.installs:
            # Attempt to parse Podman v4 and newer SQLite databases.
            if (db_path := dir.joinpath("storage/db.sql")).is_file():
                yield from self._find_containers_sqlite(db_path)

            # Fallback to globbing $PODMAN/storage/overlay-containers/* if the db does not exist (pre version 4).
            elif dir.joinpath("storage/overlay-containers").is_dir():
                yield from self._find_containers_fs(dir)

    def _find_containers_sqlite(self, path: Path) -> Iterator[PodmanContainerRecord]:
        """Find Podman containers from existing ``db.sql`` files (version 4 and newer).

        Gets info from the ``ContainerConfig`` and ``ContainerState`` tables.
        """

        try:
            db = SQLite3(path.open("rb"))
        except (ValueError, SQLError) as e:
            self.target.log.warning("Unable to read Podman database %s: %s", path, e)
            self.target.log.debug("", exc_info=e)
            return

        containers = {}

        for table_name in ["ContainerConfig", "ContainerState"]:
            if table := db.table(table_name):
                for row in table:
                    containers.setdefault(row.ID, {})
                    containers[row.ID].update(json.loads(row.JSON))

        for container_id, container in containers.items():
            volumes = []
            for mount_point in container.get("spec", {}).get("mounts", []):
                if mount_point.get("type") == "bind":
                    volumes.append(f"{mount_point.get('source')}:{mount_point.get('destination')}")  # noqa: PERF401

            # $PODMAN/storage/overlay-containers/<ID>/userdata/config.json -> `.root.path` contains the root folder,
            # this does not seem to be stored in the database at all so we have to resort to loading the `config.json`.
            try:
                config_path = path.parent.joinpath("overlay-containers", container_id, "userdata", "config.json")
                config = json.loads(config_path.read_text())
                mount_path = config.get("root", {}).get("path")
                if not mount_path:
                    raise ValueError(f"No root path found in {config_path}")  # noqa: TRY301
                mount_path = mount_path.replace("/merged", "")

            except Exception as e:
                self.target.log.warning("Unable to determine container mount path for %s: %s", container_id, e)
                config_path = None
                mount_path = None

            yield PodmanContainerRecord(
                container_id=container_id,
                image=container.get("rootfsImageName"),
                image_id=container.get("rootfsImageID"),
                command=" ".join(container.get("command", [])),
                created=container.get("createdTime"),
                running=ContainerState(container.get("state")) == ContainerState.RUNNING,
                pid=container.get("pid"),
                started=container.get("startedTime"),
                finished=container.get("finishedTime"),
                ports=list(convert_ports(container.get("newPortMappings", []))),  # TODO: research "exposedPorts"
                names=container.get("name"),
                volumes=volumes,
                environment=container.get("spec", {}).get("process", {}).get("env", []),
                mount_path=mount_path,
                config_path=config_path,
                image_path=path.parent.joinpath("overlay-images", container.get("rootfsImageID"))
                if container.get("rootfsImageID")
                else None,
                source=path,
                _target=self.target,
            )

    def _find_containers_fs(self, path: Path) -> Iterator[PodmanContainerRecord]:
        """Find Podman containers based on the ``$PODMAN/storage/overlay-containers/containers.json`` file."""

        containers = {}
        if (containers_file := path.joinpath("storage/overlay-containers/containers.json")).is_file():
            containers = json.loads(containers_file.read_text())

        for container_dir in path.joinpath("storage/overlay-containers").iterdir():
            if (
                not container_dir.is_dir()
                or not (config_path := container_dir.joinpath("userdata/config.json")).is_file()
            ):
                continue

            try:
                config = json.loads(config_path.read_text())
            except json.JSONDecodeError as e:
                self.target.log.warning("Unable to read Podman container config file %r: %r", config_path, e)
                self.target.log.debug("", exc_info=e)
                continue

            # See if this container is also registered in containers.json
            other_config = {}
            for c in containers:
                if c.get("id") == container_dir.name:
                    other_config = c
                    other_config["metadata"] = json.loads(other_config["metadata"])
                    break

            # Some fields are not available in these JSON files (we should parse BoltDB for that):
            # running state, pid, started ts, finished ts, ports, volumes.

            yield PodmanContainerRecord(
                container_id=container_dir.name,
                image=other_config.get("metadata", {}).get("image-name"),
                image_id=other_config.get("image"),
                command=" ".join(config.get("process", {}).get("args", [])),
                created=other_config.get("created"),
                names=other_config.get("names"),
                environment=config.get("process", {}).get("env", []),
                mount_path=path.joinpath(f"storage/overlay/{other_config.get('layer')}") if other_config else None,
                config_path=config_path,
                image_path=path.joinpath("storage/overlay-images", other_config.get("image")) if other_config else None,
                source=config_path,
                _target=self.target,
            )

    @export(record=PodmanLogRecord)
    def logs(self) -> Iterator[PodmanLogRecord]:
        """Returns log files (stdout/stderr) from Podman containers.

        Podman is configured by default to log towards ``syslog`` or ``journald``.
        This function parses non-default ``k8s-file`` and ``json-file`` log driver settings.

        Note that ``json-file`` is an alias of ``k8s-file`` and does not actually produce JSON output.

        Currently does not parse custom configuration in ``containers.conf`` for ``log_opt path`` values.

        References:
            - https://docs.podman.io/en/latest/markdown/podman-create.1.html#log-driver-driver
        """

        for install, _ in self.installs:
            for log_file in install.glob("storage/overlay-containers/*/userdata/ctr.log*"):
                buf = ""

                for line in log_file.open("rt"):
                    if not (match := RE_CTR_LOG.match(line)):
                        self.target.log.warning("Unable to match Podman log line %r in file %r", line, log_file)
                        continue

                    fields = match.groupdict()
                    type = fields.pop("type")

                    # Each character has it's own log line and can be concatenated up until we encounter an empty 'F'.
                    if type == "P":
                        buf += fields["message"]
                        continue
                    elif type == "F" and fields["message"] == "":
                        fields["message"] = buf
                        buf = ""

                    yield PodmanLogRecord(
                        container=log_file.parent.parent.name,
                        **fields,
                        source=log_file,
                        _target=self.target,
                    )


def convert_ports(ports: dict[str, list | dict]) -> Iterator[str]:
    for p in ports:
        yield (
            f"{p.get('host_ip') or '0.0.0.0'}:{p.get('host_port')}->"
            f"{p.get('container_port')}/{p.get('protocol', 'tcp')}"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.overlay import Overlay2Filesystem

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_overlay_filesystem_docker_container(target_linux_docker: Target) -> None:
    mount_path = target_linux_docker.fs.path(
        "/var/lib/docker/image/overlay2/layerdb/mounts/f988f88e221d97930a665712cf16ab520f7e2c5af395660c145df93aebedf071"
    )
    fs = Overlay2Filesystem(mount_path)

    assert fs.__type__ == "overlay2"
    assert len(fs.layers) == 9

    assert sorted([str(p) for p in fs.path("/").iterdir()]) == sorted(
        [
            "/home",
            "/root",
            "/media",
            "/var",
            "/tmp",
            "/mnt",
            "/bin",
            "/sys",
            "/lib",
            "/proc",
            "/dev",
            "/etc",
            "/run",
            "/sbin",
            "/opt",
            "/usr",
            "/srv",
            "/.dockerenv",
            "/container",
        ]
    )

    assert sorted([str(p) for p in fs.path("/root").iterdir()]) == [
        "/root/.ash_history",
        "/root/file.txt",
        "/root/secret.txt",
    ]

    assert fs.path("/root/secret.txt").open().read() == b"this is a secret!\n"
    assert len([str(p) for p in fs.path("/etc").iterdir()]) == 38

    # test some random symlinks
    assert fs.path("/bin/sh").is_symlink()
    assert fs.path("/bin/sh").resolve() == fs.path("/bin/busybox")
    assert fs.path("/bin/sh").readlink() == fs.path("/bin/busybox")
    assert fs.path("/etc/ssl/cert.pem").resolve() == fs.path("/etc/ssl/certs/ca-certificates.crt")
    assert fs.path("/usr/lib/libcrypto.so.3").resolve() == fs.path("/lib/libcrypto.so.3")

    # test if standard mounts were correctly added
    assert fs.path("/etc/hostname").exists()
    assert fs.path("/etc/hostname").is_file()
    assert fs.path("/etc/hostname").read_text() == "f988f88e221d\n"
    assert fs.path("/etc/hosts").exists()
    assert fs.path("/etc/resolv.conf").exists()

    # test if custom mounts were correctly added
    assert fs.path("/container/file.txt").exists()
    assert fs.path("/container/file.txt").is_file()
    assert fs.path("/container/file.txt").read_text() == "this is a mounted file!\n"
    assert fs.path("/container/folder").exists()
    assert fs.path("/container/folder").is_dir()
    assert fs.path("/container/folder/some-file.txt").exists()
    assert fs.path("/container/folder/some-file.txt").is_file()



#!/usr/bin/env python
from __future__ import annotations

import argparse
import logging
import shutil
import sys

from dissect.util.stream import RangeStream

from dissect.target.exceptions import TargetError
from dissect.target.target import Target
from dissect.target.tools.utils import (
    catch_sigpipe,
    configure_generic_arguments,
    process_generic_arguments,
)

log = logging.getLogger(__name__)
logging.lastResort = None
logging.raiseExceptions = False


@catch_sigpipe
def main() -> int:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )
    parser.add_argument("target", metavar="TARGET", help="Target to dd from")
    parser.add_argument("-w", "--write", default="-", help="output file")
    parser.add_argument("-o", "--offset", type=int, default=0, help="offset to read from.")
    parser.add_argument("-b", "--bytes", type=int, default=-1, help="amount of bytes to read")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(args)

    try:
        t = Target.open(args.target)
    except TargetError as e:
        log.error(e)  # noqa: TRY400
        log.debug("", exc_info=e)
        return 1

    if len(t.disks) > 1:
        log.error("Target has more than one disk")
        return 1

    if not len(t.disks):
        log.error("Target has no disks")
        return 1

    fhin = t.disks[0]

    fhout = sys.stdout.buffer if args.write == "-" else open(args.write, "wb")  # noqa: PTH123, SIM115

    try:
        size = args.bytes if args.bytes != -1 else fhin.size
        fhin = RangeStream(fhin, offset=args.offset, size=size)

        shutil.copyfileobj(fhin, fhout)
    finally:
        # We should not close the stdout buffer
        if fhout is not sys.stdout.buffer:
            fhout.close()

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

import fnmatch
import re
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.eventlog import evt

from dissect.target.exceptions import (
    FilesystemError,
    PluginError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
    UnsupportedPluginError,
)
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from flow.record import Record

re_illegal_characters = re.compile(r"[\(\): \.\-#]")

EvtRecordDescriptor = TargetRecordDescriptor(
    "filesystem/windows/evt",
    [
        ("datetime", "ts"),
        ("datetime", "TimeGenerated"),
        ("datetime", "TimeWritten"),
        ("string", "SourceName"),
        ("varint", "EventID"),
        ("varint", "EventCode"),
        ("varint", "EventFacility"),
        ("varint", "EventCustomerFlag"),
        ("varint", "EventSeverity"),
        ("varint", "EventType"),
        ("varint", "EventCategory"),
        ("string", "Computername"),
        ("string", "UserSid"),
        ("string[]", "Strings"),
        ("bytes", "Data"),
    ],
)


EVT_GLOB = "*.evt"


class WindowsEventlogsMixin:
    EVENTLOG_REGISTRY_KEY = "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Eventlog"
    LOGS_DIR_PATH = None

    def _get_paths(self) -> Iterator[Path]:
        seen = self.get_logs_from_dir(self.LOGS_DIR_PATH)
        yield from seen

        if self.EVENTLOG_REGISTRY_KEY:
            for reg_path in self.get_logs_from_registry():
                # We can't filter duplicates on file path alone, since "sysvol" and "C:" can show up interchangeably.
                try:
                    if any(fpath.samefile(reg_path) for fpath in seen):
                        continue
                except FilesystemError:
                    pass

                yield reg_path

    def get_logs(self, filename_glob: str = "*") -> list[Path]:
        re_filename = re.compile(fnmatch.translate(filename_glob), re.IGNORECASE)
        return [path for path in self.get_paths() if re_filename.match(str(path))]

    def get_logs_from_dir(self, logs_dir: str, filename_glob: str = "*") -> list[Path]:
        if (path := self.target.resolve(logs_dir)).exists():
            file_paths = list(path.glob(filename_glob))

            self.target.log.debug("Log files found in '%s': %d", self.LOGS_DIR_PATH, len(file_paths))
            return file_paths

        return []

    def get_logs_from_registry(self, filename_glob: str = "*") -> list[Path]:
        file_paths = []

        try:
            subkeys = self.target.registry.key(self.EVENTLOG_REGISTRY_KEY).subkeys()
        except RegistryKeyNotFoundError:
            self.target.log.warning("No eventlog registry key %s found", self.EVENTLOG_REGISTRY_KEY)
            return []
        except PluginError:
            self.target.log.warning("Cannot access registry in target")
            return []

        for subkey in subkeys:
            try:
                subkey_value = subkey.value("File")
            except RegistryValueNotFoundError:
                continue

            # resolve aliases (like `%systemroot%`) in the paths
            file_paths.append(self.target.resolve(subkey_value.value))

        return file_paths

    def check_compatible(self) -> None:
        if not self.target.resolve(self.LOGS_DIR_PATH).exists():
            raise UnsupportedPluginError(f'Event log directory "{self.LOGS_DIR_PATH}" not found')


class EvtPlugin(WindowsEventlogsMixin, Plugin):
    """Windows ``.evt`` event log plugin."""

    LOGS_DIR_PATH = "%windir%/system32/config"

    NEEDLE = b"LfLe"
    CHUNK_SIZE = 0x10000

    @arg("--logs-dir", help="logs directory to scan")
    @arg("--log-file-glob", default=EVT_GLOB, help="glob pattern to match a log file name")
    @export(record=EvtRecordDescriptor)
    def evt(self, log_file_glob: str = EVT_GLOB, logs_dir: str | None = None) -> Iterator[EvtRecordDescriptor]:
        """Parse Windows Eventlog files (``*.evt``).

        Yields dynamically created records based on the fields in the event.
        At least contains the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The TimeCreated_SystemTime field of the event.
            Provider_Name (string): The Provider_Name field of the event.
            EventID (int): The EventID of the event.
        """
        if logs_dir:
            log_paths = self.get_logs_from_dir(logs_dir, filename_glob=log_file_glob)
        else:
            log_paths = self.get_logs(filename_glob=log_file_glob)

        for entry in log_paths:
            if not entry.exists():
                self.target.log.warning("Event log file does not exist: %s", entry)
                continue

            try:
                entry_data = entry.open()
            except FilesystemError:
                self.target.log.exception("Failed to open event log: %s", entry)
                continue

            for record in evt.Evt(entry_data):
                yield self._build_record(record)

    def _build_record(self, record: Any) -> Record:
        return EvtRecordDescriptor(
            ts=record.TimeGenerated,
            TimeGenerated=record.TimeGenerated,
            TimeWritten=record.TimeWritten,
            SourceName=record.SourceName,
            EventID=record.EventID,
            EventCode=record.EventCode,
            EventFacility=record.EventFacility,
            EventCustomerFlag=record.EventCustomerFlag,
            EventSeverity=record.EventSeverity,
            EventType=record.EventType,
            EventCategory=record.EventCategory,
            Computername=record.Computername,
            Strings=record.Strings,
            Data=record.Data,
            _target=self.target,
        )

    @export(record=EvtRecordDescriptor)
    def scraped_evt(self) -> Iterator[EvtRecordDescriptor]:
        """Yields EVT log file records scraped from target disks."""
        yield from self.target.scrape.scrape_chunks_from_disks(
            needle=self.NEEDLE,
            chunk_size=self.CHUNK_SIZE,
            chunk_parser=self._parse_chunk,
            chunk_reader=self._read_chunk,
        )

    def _read_chunk(self, fh: BinaryIO, needle: bytes, offset: int, chunk_size: int) -> bytes:
        # Needle is a 2nd field in EVT header, so we need to back up
        # by 4 bytes to get to the start of the the record.
        fh.seek(offset - 4)
        return fh.read(chunk_size)

    def _parse_chunk(self, needle: bytes, chunk: bytes) -> Iterator[Record]:
        for record in evt.parse_chunk(chunk):
            yield self._build_record(record)



from __future__ import annotations

import struct
import textwrap
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO
from unittest.mock import Mock, patch

import pytest

from dissect.target import container
from dissect.target.containers import raw, vhd
from dissect.target.exceptions import ContainerError
from dissect.target.plugin import load_modules_from_paths

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def mocked_ewf_detect() -> Iterator[Mock]:
    mocked_ewf = Mock()
    mocked_ewf.EwfContainer.detect.return_value = True
    with patch.object(container, "CONTAINERS", [mocked_ewf.EwfContainer]):
        yield mocked_ewf.EwfContainer


@pytest.mark.parametrize(
    ("path", "expected_output"),
    [
        ("hello", Path("hello")),
        (["hello"], [Path("hello")]),
        ([Path("hello")], [Path("hello")]),
    ],
)
def test_open_inputs(mocked_ewf_detect: Mock, path: str | list[str] | Path, expected_output: Path | list[Path]) -> None:
    container.open(path)
    mocked_ewf_detect.assert_called_with(expected_output)


def test_open_fallback_fh(tmp_path: Path) -> None:
    # Create a valid VHD file
    fake_vhd = (
        (bytes(range(256)) * 2)
        + b"conectix"
        + (b"\x00" * 8)
        + (b"\xff" * 8)
        + (b"\x00" * 24)
        + struct.pack(">Q", 512)
        + (b"\x00" * 455)
    )

    tmp_with_ext = tmp_path.joinpath("testfile.vhd")
    tmp_without_ext = tmp_path.joinpath("testfile")
    tmp_with_wrong_ext = tmp_path.joinpath("testfile.qcow2")

    for path in [tmp_with_ext, tmp_without_ext, tmp_with_wrong_ext]:
        path.write_bytes(fake_vhd)

        assert isinstance(container.open(path), vhd.VhdContainer)
        assert vhd.VhdContainer.detect(path)

        with path.open("rb") as fh:
            assert isinstance(container.open(fh), vhd.VhdContainer)

        with path.open("rb") as fh:
            assert vhd.VhdContainer.detect(fh)

    tmp_nonexistent = tmp_path.joinpath("doesntexist")
    with pytest.raises(ContainerError):
        container.open(tmp_nonexistent)

    assert not vhd.VhdContainer.detect(tmp_nonexistent)

    tmp_dummy = tmp_path.joinpath("testdummy")
    tmp_dummy.write_bytes(b"\x00" * 1024)
    assert isinstance(container.open(tmp_dummy), raw.RawContainer)
    assert not vhd.VhdContainer.detect(tmp_dummy)


def test_reset_file_position() -> None:
    fh = BytesIO(b"\x00" * 8192)
    fh.seek(512)

    class MockContainer(container.Container):
        def __init__(self, fh: BinaryIO):
            assert fh.tell() == 0
            fh.seek(1024)
            self.success = True

        @staticmethod
        def _detect_fh(fh: BinaryIO, *args, **kwargs) -> bool:
            assert fh.tell() == 0
            fh.seek(256)
            return True

    mock_container = Mock()
    mock_container.MockContainer = MockContainer
    with patch.object(container, "CONTAINERS", [mock_container.MockContainer]):
        opened_container = container.open(fh)
        assert isinstance(opened_container, mock_container.MockContainer)
        assert opened_container.success
        assert fh.tell() == 512


def test_registration(tmp_path: Path) -> None:
    code = """
        from __future__ import annotations

        import io
        from pathlib import Path
        from typing import BinaryIO

        from dissect.target.container import Container, register


        class TestContainer(Container):
            def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
                super().__init__(fh, 20, *args, **kwargs)

            def __repr__(self) -> str:
                return f"<Container type={self.__type__} size={self.size} vs={self.vs.__type__ if self.vs else None}>"

            @staticmethod
            def _detect_fh(fh: BinaryIO, original: list[BinaryIO] | BinaryIO) -> bool:
                return False

            @staticmethod
            def detect_path(path: Path, original: list[Path] | Path) -> bool:
                return False

            def read(self, length: int) -> bytes:
                return self.fh.read(length)

            def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
                return self.fh.seek(offset, whence)

            def seekable(self) -> bool:
                return True

            def tell(self) -> int:
                return self.fh.tell()

            def close(self) -> None:
                pass


        register(__name__, TestContainer.__name__, internal=False)
    """

    (tmp_path / "container.py").write_text(textwrap.dedent(code))

    with patch("dissect.target.container.CONTAINERS", []) as mocked_containers:
        load_modules_from_paths([tmp_path])

        assert len(mocked_containers) == 1
        assert mocked_containers[0].__name__ == "TestContainer"



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.plugins.os.windows.log.etl import EtlPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.fixture
def etl_plugin(target_win: Target) -> EtlPlugin:
    return EtlPlugin(target_win)


def test_etl_plugin_shutdown(etl_plugin: EtlPlugin) -> None:
    with patch.object(etl_plugin, "read_etl_files", autospec=True):
        list(etl_plugin.shutdown())
        etl_plugin.read_etl_files.assert_called_with(etl_plugin.PATHS["shutdown"])


def test_etl_plugin_boot(etl_plugin: EtlPlugin) -> None:
    with patch.object(etl_plugin, "read_etl_files", autospec=True):
        list(etl_plugin.boot())
        etl_plugin.read_etl_files.assert_called_with(etl_plugin.PATHS["boot"])



from __future__ import annotations

import hashlib
from functools import cached_property
from typing import TYPE_CHECKING

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    from Crypto.Cipher import AES, ARC4, DES

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False


LSASecretRecord = TargetRecordDescriptor(
    "windows/credential/lsa",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "value"),
    ],
)


class LSAPlugin(Plugin):
    """Windows Local Security Authority (LSA) plugin.

    References:
        - https://learn.microsoft.com/en-us/windows/win32/secauthn/lsa-authentication
        - https://moyix.blogspot.com/2008/02/decrypting-lsa-secrets.html (Windows XP)
        - https://github.com/fortra/impacket/blob/master/impacket/examples/secretsdump.py
        - ReVaulting decryption and opportunities SANS Summit Prague 2015
    """

    __namespace__ = "lsa"

    SECURITY_POLICY_KEY = "HKEY_LOCAL_MACHINE\\SECURITY\\Policy"
    SYSTEM_KEY = "HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\LSA"

    def check_compatible(self) -> None:
        if not HAS_CRYPTO:
            raise UnsupportedPluginError("Missing pycryptodome dependency")

        if not self.target.has_function("registry") or not list(self.target.registry.keys(self.SYSTEM_KEY)):
            raise UnsupportedPluginError("Registry key not found: %s", self.SYSTEM_KEY)

    @cached_property
    def syskey(self) -> bytes:
        """Return byte value of Windows system SYSKEY, also called BootKey."""
        lsa = self.target.registry.key(self.SYSTEM_KEY)
        syskey_keys = ["JD", "Skew1", "GBG", "Data"]
        # This magic value rotates the order of the data
        alterator = [0x8, 0x5, 0x4, 0x2, 0xB, 0x9, 0xD, 0x3, 0x0, 0x6, 0x1, 0xC, 0xE, 0xA, 0xF, 0x7]

        r = bytes.fromhex("".join([lsa.subkey(key).class_name for key in syskey_keys]))
        return bytes(r[i] for i in alterator)

    @cached_property
    def lsakey(self) -> bytes:
        """Decrypt and return the LSA key of the Windows system."""
        security_pol = self.target.registry.key(self.SECURITY_POLICY_KEY)

        try:
            # Windows Vista or newer
            enc_key = security_pol.subkey("PolEKList").value("(Default)").value
            lsa_key = _decrypt_aes(enc_key, self.syskey)
            return lsa_key[68:100]
        except RegistryKeyNotFoundError:
            pass

        try:
            # Windows XP
            enc_key = security_pol.subkey("PolSecretEncryptionKey").value("(Default)").value
            lsa_key = _decrypt_rc4(enc_key, self.syskey)
            return lsa_key[16:32]
        except RegistryKeyNotFoundError:
            pass

        raise ValueError("Unable to determine LSA policy key location in registry")

    @cached_property
    def _secrets(self) -> dict[str, bytes] | None:
        """Return dict of Windows system decrypted LSA secrets.

        Includes current values (``CurrVal``) and the previous value (``OldVal``).
        Key names are suffixed with ``_OldVal`` if an old value is found in the registry.
        """
        if not self.target.ntversion:
            raise ValueError("Unable to determine Windows NT version")

        result = {}
        for subkey in self.target.registry.key(self.SECURITY_POLICY_KEY).subkey("Secrets").subkeys():
            for val in ["CurrVal", "OldVal"]:
                try:
                    enc_data = subkey.subkey(val).value("(Default)").value
                except RegistryKeyNotFoundError:
                    continue

                # Windows Vista or newer
                if float(self.target.ntversion) >= 6.0:
                    secret = _decrypt_aes(enc_data, self.lsakey)

                # Windows XP
                else:
                    secret = _decrypt_des(enc_data, self.lsakey)

                result[f"{subkey.name}{'_OldVal' if val == 'OldVal' else ''}"] = secret

        return result

    @export(record=LSASecretRecord)
    def secrets(self) -> Iterator[LSASecretRecord]:
        """Yield decrypted LSA secrets from a Windows target."""
        for key, value in self._secrets.items():
            yield LSASecretRecord(
                ts=self.target.registry.key(f"{self.SECURITY_POLICY_KEY}\\Secrets\\{key.replace('_OldVal', '')}").ts,
                name=key,
                value=value.hex(),
                _target=self.target,
            )


def _decrypt_aes(data: bytes, key: bytes) -> bytes:
    ctx = hashlib.sha256()
    ctx.update(key)
    for _ in range(1, 1000 + 1):
        ctx.update(data[28:60])

    ciphertext = data[60:]
    plaintext = []

    for i in range(0, len(ciphertext), 16):
        cipher = AES.new(ctx.digest(), AES.MODE_CBC, iv=b"\x00" * 16)
        plaintext.append(cipher.decrypt(ciphertext[i : i + 16].ljust(16, b"\x00")))

    return b"".join(plaintext)


def _decrypt_rc4(data: bytes, key: bytes) -> bytes:
    md5 = hashlib.md5()
    md5.update(key)
    for _ in range(1000):
        md5.update(data[60:76])
    rc4_key = md5.digest()

    cipher = ARC4.new(rc4_key)
    return cipher.decrypt(data[12:60])


def _decrypt_des(data: bytes, key: bytes) -> bytes:
    plaintext = []

    enc_size = int.from_bytes(data[:4], "little")
    data = data[len(data) - enc_size :]

    key0 = key
    for _ in range(0, len(data), 8):
        ciphertext = data[:8]
        block_key = _transform_key(key0[:7])

        cipher = DES.new(block_key, DES.MODE_ECB)
        plaintext.append(cipher.decrypt(ciphertext))

        key0 = key0[7:]
        data = data[8:]

        if len(key0) < 7:
            key0 = key[len(key0) :]

    return b"".join(plaintext)


def _transform_key(key: bytes) -> bytes:
    new_key = []
    new_key.append(((key[0] >> 0x01) << 1) & 0xFE)
    new_key.extend(
        (((key[i] & ((1 << (i + 1)) - 1)) << (6 - i) | (key[i + 1] >> (i + 2))) << 1) & 0xFE for i in range(6)
    )
    new_key.append(((key[6] & 0x7F) << 1) & 0xFE)
    return bytes(new_key)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

from flow.record.fieldtypes import datetime as dt

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.regf.usb import UsbPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_windows_usb(target_win_users: Target, hive_hklm: VirtualHive, hive_hku: VirtualHive) -> None:
    """Test discovery of windows usb connection history."""

    usbstor_name = "SYSTEM\\ControlSet001\\Enum\\USBSTOR\\Disk&Ven_SanDisk&Prod_Ultra&Rev_1.00"
    usbstor_key = VirtualKey(hive_hklm, usbstor_name)

    serial_name = "0401399b179e9d46555d8aa9d8618cd0a4f82512595aaea56a74bcd1c387638"
    serial_key = VirtualKey(hive_hklm, serial_name)
    serial_key.add_value(
        "ContainerID", VirtualValue(hive_hklm, "ContainerID", "{3bceaa2e-d325-5328-ba38-d55baacd43d9}")
    )
    serial_key.add_value("FriendlyName", VirtualValue(hive_hklm, "FriendlyName", "SanDisk Ultra USB Device"))
    serial_key.add_subkey("Device Parameters", VirtualKey(hive_hklm, "Device Parameters"))

    properties_key = VirtualKey(hive_hklm, "Properties")
    guid_key = VirtualKey(hive_hklm, "{83da6326-97a6-4088-9453-a1923f573b29}")

    key_0064 = VirtualKey(hive_hklm, "0064")
    key_0064.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", b"\xea\x05+$T\xdf\xda\x01"))

    key_0065 = VirtualKey(hive_hklm, "0065")
    key_0065.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", b"\xea\x05+$T\xdf\xda\x01"))

    key_0066 = VirtualKey(hive_hklm, "0066")
    key_0066.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", b"\xea\x05+$T\xdf\xda\x01"))

    key_0067 = VirtualKey(hive_hklm, "0067")
    key_0067.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", b"\x8b\xe0\x10\x96T\xdf\xda\x01"))

    guid_key.add_subkey("0064", key_0064)
    guid_key.add_subkey("0065", key_0065)
    guid_key.add_subkey("0066", key_0066)
    guid_key.add_subkey("0067", key_0067)
    properties_key.add_subkey("{83da6326-97a6-4088-9453-a1923f573b29}", guid_key)
    serial_key.add_subkey("Properties", properties_key)
    usbstor_key.add_subkey(serial_name, serial_key)
    hive_hklm.map_key(usbstor_name, usbstor_key)

    mounted_devices = VirtualKey(hive_hklm, "SYSTEM\\MountedDevices")
    mounted_devices.add_value(
        "\\DosDevices\\E:", VirtualValue(hive_hklm, "\\DosDevices\\E:", serial_name.encode("utf-16-le"))
    )
    mounted_devices.add_value(
        "\\??\\Volume{4be8862a-4b47-11ef-9a61-70d823df2914}",
        VirtualValue(hive_hklm, "\\??\\Volume{4be8862a-4b47-11ef-9a61-70d823df2914}", serial_name.encode("utf-16-le")),
    )
    hive_hklm.map_key("SYSTEM\\MountedDevices", mounted_devices)

    mounted_volumes_name = "SOFTWARE\\Microsoft\\Windows Portable Devices\\Devices\\SWD#WPDBUSENUM#_??_USBSTOR#DISK&VEN_SANDISK&PROD_ULTRA&REV_1.00#0401399B179E9D46555D8AA9D8618CD0A4F82512595AAEA56A74BCD1C387638#{53F56307-B6BF-11D0-94F2-00A0C91EFB8B}"  # noqa: E501
    mounted_volumes = VirtualKey(hive_hklm, mounted_volumes_name)
    mounted_volumes.add_value("FriendlyName", VirtualValue(hive_hklm, "FriendlyName", "Example USB"))
    hive_hklm.map_key(mounted_volumes_name, mounted_volumes)

    mountpoints2_name = (
        "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Mountpoints2\\{4be8862a-4b47-11ef-9a61-70d823df2914}"
    )
    mountpoints2 = VirtualKey(hive_hku, mountpoints2_name)
    hive_hku.map_key(mountpoints2_name, mountpoints2)

    target_win_users.add_plugin(UsbPlugin)

    with patch("dissect.target.plugins.os.windows.registry.RegistryPlugin.get_user_details") as mock_reg_user_details:
        mock_reg_user_details.return_value = target_win_users.user_details.find(username="John")
        results = list(target_win_users.usb())

    assert len(results) == 1
    assert results[0].type == "Disk"
    assert results[0].serial == "0401399b179e9d46555d8aa9d8618cd0a4f82512595aaea56a74bcd1c387638"
    assert results[0].container_id == "{3bceaa2e-d325-5328-ba38-d55baacd43d9}"
    assert results[0].vendor == "SanDisk"
    assert results[0].product == "Ultra"
    assert results[0].revision == "1.00"
    assert results[0].friendly_name == "SanDisk Ultra USB Device"
    assert results[0].first_insert == dt("2024-07-26 12:05:43.789719+00:00")
    assert results[0].first_install == dt("2024-07-26 12:05:43.789719+00:00")
    assert results[0].last_insert == dt("2024-07-26 12:05:43.789719+00:00")
    assert results[0].last_removal == dt("2024-07-26 12:08:54.878632+00:00")
    assert results[0].volumes == ["Example USB"]
    assert results[0].mounts == ["E:", "\\??\\Volume{4be8862a-4b47-11ef-9a61-70d823df2914}"]
    assert results[0].users == ["John"]
    assert results[0].source == "HKLM\\SYSTEM\\CurrentControlSet\\Enum\\USBSTOR"



from __future__ import annotations

import datetime
import gzip
import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.bsd.citrix.history import CitrixCommandHistoryPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_netscaler_bash_history_absolute_path_glob(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    commandhistory_data = """\
    Aug  9 12:56:00 <local7.notice> ns bash[23841]: root on /dev/pts/0 shell_command="find . -name '*ci.php*'"
    Aug 10 11:57:39 <local7.notice> ns bash[56440]: (null) on /dev/pts/1 shell_command="debug "hello world""
    """

    fs_bsd.map_file_fh(
        "/var/log/bash.log",
        BytesIO(textwrap.dedent(commandhistory_data).encode()),
    )

    target_citrix.add_plugin(CitrixCommandHistoryPlugin)

    results = list(target_citrix.commandhistory())
    assert len(results) == 2
    # Due to the usage of year_rollover_helper, the results are returned back-to-front. Moreover, it takes the
    # year from the file's mtime, which in this mocked version is epoch 0 (thus the year 1970)
    assert results[1].ts == datetime.datetime(1970, 8, 9, 12, 56, 0, tzinfo=datetime.timezone.utc)
    assert results[1].command == "find . -name '*ci.php*'"
    assert results[1].order == -1
    assert results[1].shell == "citrix-netscaler-bash"
    assert results[1].source.as_posix() == "/var/log/bash.log"

    assert results[0].ts == datetime.datetime(1970, 8, 10, 11, 57, 39, tzinfo=datetime.timezone.utc)
    assert results[0].command == 'debug "hello world"'
    assert results[0].order == 0
    assert results[0].shell == "citrix-netscaler-bash"
    assert results[0].source.as_posix() == "/var/log/bash.log"


def test_netscaler_commandhistory_decompress(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    commandhistory_data = """\
    Aug  9 12:56:00 <local7.notice> ns bash[23841]: root on /dev/pts/0 shell_command="find . -name '*ci.php*'"
    Aug 10 11:57:39 <local7.notice> ns bash[56440]: (null) on /dev/pts/1 shell_command="debug "hello world""
    """

    fs_bsd.map_file_fh(
        "/var/log/bash.log.0.gz",
        BytesIO(gzip.compress(textwrap.dedent(commandhistory_data).encode())),
    )

    target_citrix.add_plugin(CitrixCommandHistoryPlugin)

    results = list(target_citrix.commandhistory())
    assert len(results) == 2
    # Due to the usage of year_rollover_history, the results are returned back-to-front
    assert results[1].ts == datetime.datetime(1970, 8, 9, 12, 56, 0, tzinfo=datetime.timezone.utc)
    assert results[1].command == "find . -name '*ci.php*'"
    assert results[1].order == -1
    assert results[1].shell == "citrix-netscaler-bash"
    assert results[1].source.as_posix() == "/var/log/bash.log.0.gz"

    assert results[0].ts == datetime.datetime(1970, 8, 10, 11, 57, 39, tzinfo=datetime.timezone.utc)
    assert results[0].command == 'debug "hello world"'
    assert results[0].order == 0
    assert results[0].shell == "citrix-netscaler-bash"
    assert results[0].source.as_posix() == "/var/log/bash.log.0.gz"


def test_netscaler_cli_history(target_citrix: Target, fs_bsd: VirtualFilesystem) -> None:
    commandhistory_data = """\
    _HiStOrY_V2_
    help
    shell
    """

    fs_bsd.map_file_fh("/var/nstmp/user/.nscli_history", BytesIO(textwrap.dedent(commandhistory_data).encode()))

    target_citrix.add_plugin(CitrixCommandHistoryPlugin)

    results = list(target_citrix.commandhistory())
    assert len(results) == 2

    assert not results[0].ts
    assert results[0].command == "help"
    assert results[0].order == 0
    assert results[0].shell == "citrix-netscaler-cli"
    assert results[0].source.as_posix() == "/var/nstmp/user/.nscli_history"

    assert not results[1].ts
    assert results[1].command == "shell"
    assert results[1].order == 1
    assert results[1].shell == "citrix-netscaler-cli"
    assert results[1].source.as_posix() == "/var/nstmp/user/.nscli_history"



from __future__ import annotations

import gzip
import tarfile
import textwrap
from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch
from zoneinfo import ZoneInfo

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.log.messages import MessagesPlugin, MessagesRecord
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem


def test_unix_log_messages_plugin(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))

    data_file = absolute_path("_data/plugins/os/unix/log/messages/messages")
    fs_unix.map_file("var/log/messages", data_file)

    entry = fs_unix.get("var/log/messages")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix_users.add_plugin(MessagesPlugin)

        results = list(target_unix_users.messages())
        assert len(results) == 2

        assert results[0].ts == datetime(2022, 1, 1, 13, 21, 34, tzinfo=ZoneInfo("Europe/Amsterdam"))
        assert results[0].message == "Stopped target Swap."
        assert results[0].pid is None
        assert results[0].source == "/var/log/messages"

        assert results[1].ts == datetime(2021, 12, 31, 3, 14, 15, tzinfo=ZoneInfo("Europe/Amsterdam"))
        assert results[1].message == "Starting Journal Service..."
        assert results[1].pid == 1
        assert results[1].source == "/var/log/messages"

    # assure syslog() behaves the same as messages()
    syslogs = list(target_unix_users.syslog())
    assert len(syslogs) == len(results)
    assert isinstance(syslogs[0], type(MessagesRecord()))


def test_unix_log_messages_compressed_timezone_year_rollover() -> None:
    target = Target()
    bio = BytesIO()

    with tarfile.open(mode="w:gz", fileobj=bio) as tf:
        # UTC-6 tar filesystem
        timezone_contents = "America/Chicago"
        timezone_file = BytesIO(textwrap.dedent(timezone_contents).encode())
        timezone_tar_info = tarfile.TarInfo("etc/timezone")
        timezone_tar_info.size = len(timezone_contents)
        tf.addfile(timezone_tar_info, timezone_file)

        # Create a tar file with correct mtime
        messages_log_contents = """\
        Dec 31 03:14:00 localhost systemd[1]: Starting Journal Service...
        \xfean  1 03:14:00 localhost systemd[1]: Starting Journal Service...
        Jan  1 13:37:00 localhost systemd: Stopped target Swap."""
        messages_log_file = BytesIO(textwrap.dedent(messages_log_contents).encode())
        messages_log_tar_info = tarfile.TarInfo("var/log/messages")
        messages_log_tar_info.size = len(textwrap.dedent(messages_log_contents))
        messages_log_tar_info.mtime = 1640995800  # (Sat 1 January 2022 00:10:00 UTC)
        tf.addfile(messages_log_tar_info, messages_log_file)

    fs = TarFilesystem(bio)
    target.filesystems.add(fs)
    target.fs.mount("/", fs)
    target._os_plugin = UnixPlugin
    target.apply()
    target.add_plugin(MessagesPlugin)

    results = list(target.messages())
    results.reverse()

    assert len(results) == 2
    assert isinstance(results[0], type(MessagesRecord()))
    assert isinstance(results[1], type(MessagesRecord()))
    assert results[0].ts == datetime(2020, 12, 31, 3, 14, 0, tzinfo=ZoneInfo("America/Chicago"))
    assert results[1].ts == datetime(2021, 1, 1, 13, 37, 0, tzinfo=ZoneInfo("America/Chicago"))


def test_unix_log_messages_malformed_log_year_rollover(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))

    messages = BytesIO(
        b"Dec 31 03:14:00 localhost systemd[1]: Starting Journal Service...\r\n"
        b"\xefan  1 13:37:00 localhost systemd: Stopped target Swap.\r\n"
        b"Dec 31 03:14:00 localhost systemd[1]: Starting Journal Service...\r\n"
    )
    fs_unix.map_file_fh("var/log/messages", messages)

    entry = fs_unix.get("var/log/messages")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix_users.add_plugin(MessagesPlugin)

        results = list(target_unix_users.messages())
        assert len(results) == 2

        assert results[0].ts
        assert results[0].service == "systemd"
        assert results[0].pid == 1
        assert results[0].message == "Starting Journal Service..."
        assert results[0].source == "/var/log/messages"


def test_unix_messages_cloud_init(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly parse plaintext and compressed cloud-init log files."""

    messages = """
    2005-08-09 11:55:21,000 - foo.py[DEBUG]: This is a cloud-init message!
    2005-08-09 11:55:21,001 - util.py[DEBUG]: Cloud-init v. 1.2.3-4ubuntu5 running 'init-local' at Tue, 9 Aug 2005 11:55:21 +0000. Up 13.37 seconds.
    """  # noqa: E501
    msg_bytes = textwrap.dedent(messages).encode()

    fs_unix.map_file_fh("/etc/timezone", BytesIO(b"Europe/Amsterdam"))
    fs_unix.map_file_fh("/var/log/installer/cloud-init.log", BytesIO(msg_bytes))
    fs_unix.map_file_fh("/var/log/installer/cloud-init.log.1.gz", BytesIO(gzip.compress(msg_bytes)))
    target_unix.add_plugin(MessagesPlugin)

    results = sorted(target_unix.messages(), key=lambda r: r.source)
    assert len(results) == 4

    assert results[0].ts == datetime(2005, 8, 9, 11, 55, 21, 0, tzinfo=ZoneInfo("Europe/Amsterdam"))
    assert results[0].service == "foo.py"
    assert results[0].pid is None
    assert results[0].message == "This is a cloud-init message!"
    assert results[0].source == "/var/log/installer/cloud-init.log"

    assert results[-1].ts == datetime(2005, 8, 9, 11, 55, 21, 1_000, tzinfo=ZoneInfo("Europe/Amsterdam"))
    assert results[-1].service == "util.py"
    assert results[-1].pid is None
    assert (
        results[-1].message
        == "Cloud-init v. 1.2.3-4ubuntu5 running 'init-local' at Tue, 9 Aug 2005 11:55:21 +0000. Up 13.37 seconds."
    )
    assert results[-1].source == "/var/log/installer/cloud-init.log.1.gz"


def test_unix_messages_ts_iso_8601_format(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly detect and parse ISO 8601 formatted syslog logs."""

    fs_unix.map_file_fh("/etc/hostname", BytesIO(b"hostname"))
    messages = """
    2024-12-31T13:37:00.123456+02:00 hostname systemd[1]: Started anacron.service - Run anacron jobs.
    2024-12-31T13:37:00.123456+02:00 hostname anacron[1337]: Anacron 2.3 started on 2024-12-31
    2024-12-31T13:37:00.123456+02:00 hostname anacron[1337]: Normal exit (0 jobs run)
    2024-12-31T13:37:00.123456+02:00 hostname systemd[1]: anacron.service: Deactivated successfully.
    """
    fs_unix.map_file_fh("/var/log/syslog.1", BytesIO(gzip.compress(textwrap.dedent(messages).encode())))

    target_unix.add_plugin(UnixPlugin)
    target_unix.add_plugin(MessagesPlugin)
    results = sorted(target_unix.syslog(), key=lambda r: r.ts)

    assert len(results) == 4

    assert results[0].hostname == "hostname"
    assert results[0].service == "systemd"
    assert results[0].pid == 1
    assert results[0].ts == datetime(2024, 12, 31, 11, 37, 0, 123456, tzinfo=timezone.utc)
    assert results[0].message == "Started anacron.service - Run anacron jobs."
    assert results[0].source == "/var/log/syslog.1"


def test_linux_messages_kernel_logs(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can parse kernel ring buffer messages."""

    messages = """
    Dec 31 13:37:01 hostname kernel: [1337.1337] some example message
    Jan  1 13:37:02 kernel: [    0.000000] x86/fpu: Supporting feature 0x1337: 'message'
    Jan  2 13:37:03 kernel: [    0.000000] x86/fpu: state[1337]:  1337, size[1337]:   1337
    Jan  3 13:37:04 kernel:
    """

    fs_unix.map_file_fh("/var/log/installer/syslog.1", BytesIO(gzip.compress(textwrap.dedent(messages).encode())))
    target_unix.add_plugin(UnixPlugin)
    target_unix.add_plugin(MessagesPlugin)

    results = sorted(target_unix.syslog(), key=lambda r: r.ts)
    assert len(results) == 4

    assert results[0].service == "kernel"
    assert results[0].message == "[1337.1337] some example message"

    assert results[1].service == "kernel"
    assert results[1].message == "[    0.000000] x86/fpu: Supporting feature 0x1337: 'message'"

    assert results[2].service == "kernel"
    assert results[2].message == "[    0.000000] x86/fpu: state[1337]:  1337, size[1337]:   1337"

    assert results[3].service == "kernel"
    assert results[3].message is None



from __future__ import annotations

import calendar
from datetime import datetime, timedelta, timezone, tzinfo
from typing import TYPE_CHECKING, NamedTuple

from dissect.cstruct import cstruct

from dissect.target.exceptions import (
    RegistryError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
)
from dissect.target.helpers.mui import MUI_TZ_MAP
from dissect.target.plugin import internal
from dissect.target.plugins.os.default.datetime import DateTimePlugin

if TYPE_CHECKING:
    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.target import Target

tz_def = """
typedef struct _SYSTEMTIME {
    WORD wYear;
    WORD wMonth;
    WORD wDayOfWeek;
    WORD wDay;
    WORD wHour;
    WORD wMinute;
    WORD wSecond;
    WORD wMilliseconds;
} SYSTEMTIME;

typedef struct _REG_TZI_FORMAT {
    LONG Bias;
    LONG StandardBias;
    LONG DaylightBias;
    SYSTEMTIME StandardDate;
    SYSTEMTIME DaylightDate;
} REG_TZI_FORMAT;
"""
c_tz = cstruct().load(tz_def)


# Althoug calendar.SUNDAY is only officially documented since Python 3.10, it
# is present in Python 3.9, so we ignore the vermin warnings.
SUNDAY = calendar.SUNDAY  # novermin
SundayFirstCalendar = calendar.Calendar(SUNDAY)


class TimezoneInformation(NamedTuple):
    # The current bias from UTC
    bias: timedelta
    # The additional bias on top of the current bias when in standard time (usually 0)
    standard_bias: timedelta
    # The additional bias on top of the current bias when in daylight saving time (usually -60)
    daylight_bias: timedelta
    # When standard time starts, or in other words, daylight saving ends (see parse_systemtime_transition)
    standard_date: c_tz._SYSTEMTIME
    # When daylight saving starts (see parse_systemtime_transition)
    daylight_date: c_tz._SYSTEMTIME


ZERO = timedelta(0)
HOUR = timedelta(hours=1)


def parse_systemtime_transition(systemtime: c_tz._SYSTEMTIME, year: int) -> datetime:
    """Return the transition datetime for a given year using the SYSTEMTIME of a STD or DST transition date.

    The SYSTEMTIME date of a TZI structure needs to be used to calculate the actual date for a given year.
    The wMonth member indicates the month, the wDayOfWeek member indicates the weekday and the wDay indicates the
    occurance of the day of the week within the month (1 to 5, where 5 indicates the final occurrence during the
    month if that day of the week does not occur 5 times).

    Reference:
        - https://docs.microsoft.com/en-us/windows/win32/api/timezoneapi/ns-timezoneapi-time_zone_information
    """
    if not (1 <= systemtime.wDay <= 5):
        raise ValueError("systemtime.wDay should be between 1 and 5")

    month = SundayFirstCalendar.monthdayscalendar(year, systemtime.wMonth)
    occurrences = [week[systemtime.wDayOfWeek] for week in month if week[systemtime.wDayOfWeek]]
    target_occurrence = -1 if systemtime.wDay == 5 and len(occurrences) < 5 else systemtime.wDay - 1

    return datetime(  # noqa: DTZ001
        year,
        systemtime.wMonth,
        occurrences[target_occurrence],
        systemtime.wHour,
        systemtime.wMinute,
        systemtime.wSecond,
        systemtime.wMilliseconds * 1000,
    )


def parse_dynamic_dst(key: RegistryKey) -> dict[int, TimezoneInformation]:
    """Parse dynamic DST information from a given TimeZoneInformation registry key.

    If a timezone has dynamic DST information, there's a "Dynamic DST" subkey with values for each year.
    The FirstEntry and LastEntry contain the first and last year respectively. The TZI structure is the same as the
    main TimeZoneInformation TZI.
    """
    result = {}
    try:
        dynamic_key = key.subkey("Dynamic DST")
        first = int(dynamic_key.value("FirstEntry").value)
        last = int(dynamic_key.value("LastEntry").value)

        for year in range(first, last + 1):
            result[year] = parse_tzi(dynamic_key.value(str(year)).value)
    except RegistryError:
        pass

    return result


def parse_tzi(tzi: bytes) -> TimezoneInformation:
    """Parse binary TZI into a ``TimezoneInformation`` namedtuple."""
    parsed = c_tz.REG_TZI_FORMAT(tzi)
    return TimezoneInformation(
        parsed.Bias,
        parsed.StandardBias,
        parsed.DaylightBias,
        parsed.StandardDate,
        parsed.DaylightDate,
    )


def get_dst_range(tzi: TimezoneInformation, year: int) -> tuple[datetime, datetime]:
    """Get the start and end date of DST for the given year."""
    start = parse_systemtime_transition(tzi.daylight_date, year)
    end = parse_systemtime_transition(tzi.standard_date, year)

    return (start, end)


class WindowsTimezone(tzinfo):
    """A ``datetime.tzinfo`` class representing a timezone from parsed Windows TZI data.

    Mostly inspired by the examples in the Python documentation.
    """

    def __init__(self, name: str, key: RegistryKey):
        self.name = name
        self.display = translate_tz(key, "Display")
        self.dlt_name = translate_tz(key, "Dlt")
        self.std_name = translate_tz(key, "Std")
        self.tzi = parse_tzi(key.value("TZI").value)
        self.dynamic_dst = parse_dynamic_dst(key)

        self.bias = timedelta(minutes=-self.tzi.bias)

    def __repr__(self) -> str:
        return self.display

    def is_dst(self, dt: datetime) -> bool:
        assert dt.tzinfo is self

        tzi = self.dynamic_dst.get(dt.year, self.tzi)
        # If the time zone does not support daylight saving time the wMonth member is zero
        if tzi.daylight_date.wMonth == 0 or tzi.standard_date.wMonth == 0:
            return False

        start, end = get_dst_range(tzi, dt.year)

        # DST is flipped in some regions
        flip = False
        if start > end:
            flip = True
            start, end = end, start

        # Can't compare naive to aware objects, so strip the timezone from dt first
        # Cast to a proper datetime object to avoid issues with subclassed datetime objects
        dt = datetime(  # noqa: DTZ001
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            dt.microsecond,
            tzinfo=None,
            fold=dt.fold,
        )

        result = False
        if start + HOUR <= dt < end - HOUR:
            # DST is in effect.
            result = True
        elif end - HOUR <= dt < end:
            # Fold (an ambiguous hour): use dt.fold to disambiguate.
            result = not dt.fold
        elif start <= dt < start + HOUR:
            # Gap (a non-existent hour): reverse the fold rule.
            result = bool(dt.fold)
        else:
            # DST is off.
            result = False

        return not result if flip else result

    def utcoffset(self, dt: datetime) -> int:
        # Windows timezones use a default bias and a separate additional bias when in STD or DST
        return self.bias + self.dst(dt)

    def dst(self, dt: datetime) -> timedelta:
        tzi = self.dynamic_dst.get(dt.year, self.tzi)
        std_bias = timedelta(minutes=-tzi.standard_bias)
        dst_bias = timedelta(minutes=-tzi.daylight_bias)

        return dst_bias if self.is_dst(dt) else std_bias

    def tzname(self, dt: datetime) -> str:
        return self.dlt_name if self.is_dst(dt) else self.std_name


class WindowsDateTimePlugin(DateTimePlugin):
    __namespace__ = "datetime"

    def __init__(self, target: Target):
        super().__init__(target)

        tz_name = None

        try:
            timezone_information = self.target.registry.key(
                "HKLM\\SYSTEM\\CurrentControlSet\\Control\\TimeZoneInformation"
            )

            try:
                # Windows 7+
                tz_name = timezone_information.value("TimeZoneKeyName").value
            except RegistryValueNotFoundError:
                # < Windows 7 (https://github.com/dateutil/dateutil/issues/210)
                tz_name_localized = timezone_information.value("StandardName").value
                for timezone_key in self.target.registry.key(
                    "HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones"
                ).subkeys():
                    if tz_name_localized == timezone_key.value("Std").value:
                        tz_name = timezone_key.name
        except RegistryKeyNotFoundError:
            pass

        if tz_name is None:
            self.target.log.warning("Could not determine timezone of target, falling back to UTC for datetime helpers")
            self._tzinfo = timezone.utc
        else:
            self._tzinfo = self.tz(tz_name)

    def check_compatible(self) -> None:
        pass

    @internal
    def tz(self, name: str) -> tzinfo:
        """Return a ``datetime.tzinfo`` of the given timezone name."""
        tz_data = self.target.registry.key(f"HKLM\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\{name}")
        return WindowsTimezone(name, tz_data)

    @internal(property=True)
    def tzinfo(self) -> tzinfo:
        """Return a ``datetime.tzinfo`` of the current system timezone."""
        return self._tzinfo


def translate_tz(key: RegistryKey, name: str) -> str:
    """Translate a timezone resource string to English.

    Non-English distributions of Windows contain a local translation in the "Display", "Dlt" and "Std" keys.
    The ``MUI_*`` keys contain a reference to the English timezone name we want, e.g. "@tzres.dll,-1337".
    """
    try:
        string_id = int(key.value("MUI_" + name).value.replace("@tzres.dll,-", ""))
        if translation := MUI_TZ_MAP.get(string_id):
            return translation
    except (RegistryValueNotFoundError, ValueError):
        pass

    return key.value(name).value



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.overlay import Overlay2Filesystem
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


class Overlay2Loader(Loader):
    """Load Docker overlay2 filesystems."""

    @staticmethod
    def detect(path: TargetPath) -> bool:
        # path should be a folder
        if not path.is_dir():
            return False

        # with the following three files
        for required_file in ["init-id", "parent", "mount-id"]:
            if not path.joinpath(required_file).exists():
                return False

        # and should have the following parent folders
        return "image/overlay2/layerdb/mounts/" in path.as_posix()

    def map(self, target: Target) -> None:
        target.filesystems.add(Overlay2Filesystem(self.absolute_path))



from __future__ import annotations

import fnmatch
import re
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import arg, export
from dissect.target.plugins.general.config import (
    ConfigurationEntry,
    ConfigurationTreePlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

UnixConfigTreeRecord = TargetRecordDescriptor(
    "unix/config",
    [
        ("path", "source"),
        ("path", "path"),
        ("string", "key"),
        ("string[]", "value"),
    ],
)


class EtcTree(ConfigurationTreePlugin):
    """Unix etc configuration tree plugin."""

    __namespace__ = "etc"

    def __init__(self, target: Target):
        super().__init__(target, "/etc")

    def _sub(
        self, items: ConfigurationEntry | dict, entry: Path, orig_path: Path, pattern: str
    ) -> Iterator[UnixConfigTreeRecord]:
        index = 0
        if not isinstance(items, dict):
            items = items.as_dict()

        for raw_key, value in items.items():
            key = re.sub(r"[\n\r\t]", "", raw_key)
            path = Path(entry) / Path(key)

            if isinstance(value, dict):
                yield from self._sub(value, path, orig_path, pattern)
                continue

            if not isinstance(value, list):
                value = [str(value)]

            if fnmatch.fnmatch(path, pattern):
                data = {
                    "_target": self.target,
                    "source": self.target.fs.path(orig_path),
                    "path": path,
                    "key": key,
                    "value": value,
                }
                if value == [""]:
                    data["key"] = index
                    data["value"] = [key]
                    index += 1

                yield UnixConfigTreeRecord(**data)

    @export(record=UnixConfigTreeRecord)
    @arg("--glob", dest="pattern", default="*", help="Glob-style pattern to search for")
    @arg("--root", dest="root", default="/", help="Path to use as root for search")
    def etc(self, pattern: str, root: str) -> Iterator[UnixConfigTreeRecord]:
        """Yield etc configuration records."""

        for entry, _, items in self.config_fs.walk(root):
            for item in items:
                try:
                    path = Path(entry) / item
                    config_object = self.get(str(path))

                    if isinstance(config_object, ConfigurationEntry):
                        yield from self._sub(config_object, path, orig_path=path, pattern=pattern)
                except Exception:  # noqa: PERF203
                    self.target.log.warning("Could not open configuration item: %s", item)



from __future__ import annotations

import importlib
from typing import Any


class LazyImport:
    def __init__(self, module_name: str):
        self._module_name = module_name
        self._module = None
        self._loaded = False

    def __repr__(self) -> str:
        return f"<lazy {self._module_name} loaded={self._loaded} failed={isinstance(self._module, FailedImport)}>"

    def __getattr__(self, attr: str) -> LazyAttr:
        return LazyAttr(self, attr)

    def _import(self) -> None:
        if not self._module:
            try:
                self._module = importlib.import_module(self._module_name)
            except Exception as e:
                self._module = FailedImport(self, e)
            finally:
                self._loaded = True


class FailedImport:
    def __init__(self, module: LazyImport, exc: Exception):
        self._module = module
        self._exc = exc

    def __getattr__(self, attr: str) -> None:
        self._error()

    def __call__(self, *args, **kwargs) -> None:
        self._error()

    def _error(self) -> None:
        raise ImportError(f"Failed to lazily import {self._module._module_name}: {self._exc}")


class LazyAttr:
    def __init__(self, module: LazyImport, attr: str):
        self._module = module
        self._attr = attr
        self._realattr = None
        self._loaded = False
        self._exc = None

    def __repr__(self) -> str:
        return (
            f"<lazyattr {self._module._module_name}.{self._attr} loaded={self._loaded} failed={self._exc is not None}>"
        )

    @property
    def __doc__(self) -> str:
        try:
            return self._load().__doc__
        except Exception as e:
            return f"Failed to load docstring: {e}"

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._load(), attr)

    def __call__(self, *args, **kwargs) -> Any:
        return self._load()(*args, **kwargs)

    def _load(self) -> Any:
        if not self._loaded:
            self._module._import()
            try:
                self._realattr = getattr(self._module._module, self._attr)
            except Exception as e:
                self._exc = e
                raise
            finally:
                self._loaded = True

        if self._exc:
            raise self._exc

        return self._realattr


def import_lazy(module_name: str) -> LazyImport:
    return LazyImport(module_name)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.browser.edge import EdgePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target

# NOTE: Missing cookie tests for Edge.


@pytest.fixture
def target_edge_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Default\\",
        absolute_path("_data/plugins/apps/browser/edge/"),
    )
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Profile 1\\",
        absolute_path("_data/plugins/apps/browser/edge/"),
    )

    target_win_users.add_plugin(EdgePlugin)

    return target_win_users


@pytest.fixture
def target_edge_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Target:
    fs_unix.map_dir("/root/.config/microsoft-edge/Default/", absolute_path("_data/plugins/apps/browser/edge/"))
    fs_unix.map_dir("/root/.config/microsoft-edge/Profile 1/", absolute_path("_data/plugins/apps/browser/edge/"))
    target_unix_users.add_plugin(EdgePlugin)

    return target_unix_users


@pytest.fixture
def target_edge_win_snapshot(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Snapshots\\116.0.5038.150\\Default",
        absolute_path("_data/plugins/apps/browser/edge/"),
    )
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Snapshots\\116.0.5038.150\\Profile 1",
        absolute_path("_data/plugins/apps/browser/edge/"),
    )

    target_win_users.add_plugin(EdgePlugin)

    return target_win_users


@pytest.mark.parametrize(
    "target_platform",
    ["target_edge_win", "target_edge_unix", "target_edge_win_snapshot"],
)
def test_edge_history(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.edge.history())

    assert len(records) == 90
    assert {"edge"} == {record.browser for record in records}

    assert records[-1].url == "https://github.com/fox-it/dissect"
    assert records[-1].id == 45
    assert records[-1].visit_count == 2
    assert records[-1].ts == dt("2023-02-24T11:54:44.875477+00:00")


@pytest.mark.parametrize(
    "target_platform",
    ["target_edge_win", "target_edge_unix", "target_edge_win_snapshot"],
)
def test_edge_downloads(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.edge.downloads())

    assert len(records) == 4
    assert {"edge"} == {record.browser for record in records}

    assert records[0].id == 1
    assert records[0].ts_start == dt("2023-02-24T11:52:36.631304+00:00")
    assert records[0].ts_end == dt("2023-02-24T11:52:37.068768+00:00")
    assert records[0].url == "https://codeload.github.com/fox-it/dissect/zip/refs/heads/main"
    assert records[0].state == "complete"


@pytest.mark.parametrize(
    "target_platform",
    ["target_edge_win", "target_edge_unix", "target_edge_win_snapshot"],
)
def test_edge_extensions(target_platform: Target, request: pytest.FixtureRequest) -> None:
    target_platform = request.getfixturevalue(target_platform)
    records = list(target_platform.edge.extensions())

    assert len(records) == 78
    assert {"edge"} == {record.browser for record in records}

    assert records[0].ts_install == dt("2023-04-18T08:39:57.968208+00:00")
    assert records[0].ts_update == dt("2023-04-18T08:39:57.968208+00:00")
    assert records[0].name == "Web Store"
    assert records[0].version == "0.2"
    assert records[0].extension_id == "ahfgeienlihckogmohjhadlkjgocpleb"


def test_windows_edge_passwords_plugin(target_edge_win: Target) -> None:
    records = list(target_edge_win.edge.passwords())

    assert len(records) == 4

    for record in records:
        assert record.browser == "edge"
        assert record.decrypted_username == "username"
        assert record.decrypted_password is None

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_edge_passwords_basic_plugin(target_edge_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file(
        "/root/.config/microsoft-edge/Default/Login Data",
        absolute_path("_data/plugins/apps/browser/chromium/unix/basic/Login Data"),
    )
    fs_unix.map_file(
        "/root/.config/microsoft-edge/Profile 1/Login Data",
        absolute_path("_data/plugins/apps/browser/chromium/unix/basic/Login Data"),
    )

    records = list(target_edge_unix.edge.passwords())

    assert len(records) == 4

    for record in records:
        assert record.browser == "edge"
        assert record.decrypted_username == "username"
        assert record.decrypted_password == "password"

    assert records[0].url == "https://example.com/"
    assert records[1].url == "https://example.org/"


def test_unix_edge_passwords_gnome_plugin(target_edge_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file(
        "/root/.config/microsoft-edge/Default/Login Data",
        absolute_path("_data/plugins/apps/browser/chromium/unix/gnome/Login Data"),
    )
    fs_unix.map_file(
        "/root/.config/microsoft-edge/Profile 1/Login Data",
        absolute_path("_data/plugins/apps/browser/chromium/unix/gnome/Login Data"),
    )

    records = list(target_edge_unix.edge.passwords())

    assert len(records) == 2

    assert records[0].decrypted_username == "username"
    assert records[0].decrypted_password is None
    assert records[0].url == "https://test.com/"


def test_edge_windows_snapshots(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    base_dirs = [
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Default",
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Profile 1",
    ]
    snapshot_dirs = [
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Snapshots\\116.0.5038.150\\Default",
        "Users\\John\\AppData\\Local\\Microsoft\\Edge\\User Data\\Snapshots\\119.0.7845.119\\Default",
    ]
    profile_dirs = base_dirs + snapshot_dirs

    for dir in profile_dirs:
        fs_win.map_dir(
            dir,
            absolute_path("_data/plugins/apps/browser/edge/"),
        )

    target_win_users.add_plugin(EdgePlugin)

    records_list = [
        list(target_win_users.edge.history()),
        list(target_win_users.edge.extensions()),
        list(target_win_users.edge.downloads()),
    ]

    # Loop over the different types of records and verify we have the same amount of records in each profile directory.
    for records in records_list:
        assert {"edge"} == {record.browser for record in records}

        for base_dir in base_dirs:
            base_path_records = [r for r in records if str(r.source.parent).endswith(base_dir)]

        for snapshot_dir in snapshot_dirs:
            # Retrieve records that are in the snapshot's directory.
            snapshot_records = [r for r in records if str(r.source.parent).endswith(snapshot_dir)]

        # We map the same files in each of the snapshot directories.
        assert len(base_path_records) == len(snapshot_records)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

MuiCacheRecord = create_extended_descriptor([RegistryRecordDescriptorExtension, UserRecordDescriptorExtension])(
    "windows/registry/muicache",
    [
        ("varint", "index"),
        ("string", "name"),
        ("string", "value"),
        ("path", "path"),
    ],
)


class MuiCachePlugin(Plugin):
    """Plugin that iterates various MUIcache locations."""

    KEYS = (
        "HKCU\\Software\\Classes\\Local Settings\\Software\\Microsoft\\Windows\\Shell\\MuiCache",  # NT >= 6.0
        "HKCU\\Software\\Classes\\Local Settings\\MuiCache",  # NT >= 6.0
        "HKCU\\Software\\Microsoft\\Windows\\ShellNoRoam\\MUICache",  # NT < 6.0
    )

    FIELD_NAMES = ("FriendlyAppName", "ApplicationCompany")

    def check_compatible(self) -> None:
        if not next(self.target.registry.keys(self.KEYS), None):
            raise UnsupportedPluginError("No MuiCache registry keys found")

    @export(record=MuiCacheRecord)
    def muicache(self) -> MuiCacheRecord:
        """Iterate various MUIcache key locations.

        The MUIcache registry key stores information about executed GUI-based programs. The key is part of
        the Multilingual User Interface service in Windows. MUIcache references the file description contained within
        the executable's resource section and populates that value.

        References:
            - https://www.magnetforensics.com/blog/forensic-analysis-of-muicache-files-in-windows/
            - https://forensafe.com/blogs/muicache.html

        Yields MuiCacheRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            index (varint): The index of the entry.
            name (string): The value name.
            value (string): The value.
            path (path): The executable path.
        """
        for reg_path in self.KEYS:
            for key in self.target.registry.keys(reg_path):
                if len(key.subkeys()):
                    for subkey in key.subkeys():
                        for item in subkey.subkeys():
                            yield from self._get_records(item)
                else:
                    yield from self._get_records(key)

    def _get_records(self, key: RegistryKey) -> Iterator[MuiCacheRecord]:
        for index, entry in enumerate(key.values()):
            user = self.target.registry.get_user(key)
            try:
                if entry.name.endswith(self.FIELD_NAMES):
                    entry_path, name = entry.name.rsplit(".", 1)
                else:
                    name = None
                    entry_path = entry.name.rsplit(",-", 1)[0]

                # Filter on the type of the registry value
                if isinstance(entry.value, bytes):
                    continue

                yield MuiCacheRecord(
                    index=index,
                    name=name,
                    value=entry.value,
                    path=self.target.fs.path(entry_path),
                    _target=self.target,
                    _key=key,
                    _user=user,
                )
            except ValueError:
                continue
            except Exception:
                self.target.log.exception("Exception while parsing muicache")
                continue



from __future__ import annotations

import logging
import stat
import zipfile
from datetime import datetime, timezone
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import BufferedStream

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import (
    Filesystem,
    FilesystemEntry,
    VirtualDirectory,
    VirtualFilesystem,
)
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

log = logging.getLogger(__name__)


class ZipFilesystem(Filesystem):
    """Filesystem implementation for zip files.

    Python does not have symlink support in the zipfile module, so that's not currently supported.
    See https://github.com/python/cpython/issues/82102 for more information.
    """

    __type__ = "zip"

    def __init__(
        self,
        fh: BinaryIO,
        base: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(fh, *args, **kwargs)

        fh.seek(0)

        self.zip = zipfile.ZipFile(fh, mode="r")
        self.base = base or ""

        self._fs = VirtualFilesystem(alt_separator=self.alt_separator, case_sensitive=self.case_sensitive)

        for member in self.zip.infolist():
            mname = member.filename.strip("/")
            if not mname.startswith(self.base) or mname == ".":
                continue

            rel_name = self._resolve_path(mname)
            self._fs.map_file_entry(rel_name, ZipFilesystemEntry(self, rel_name, member))

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a zip file on a given file-like object."""
        return zipfile.is_zipfile(fh)

    def _resolve_path(self, path: str) -> str:
        return fsutil.normpath(path[len(self.base) :], alt_separator=self.alt_separator)

    def get(self, path: str, relentry: FilesystemEntry = None) -> FilesystemEntry:
        """Returns a ZipFilesystemEntry object corresponding to the given path."""
        return self._fs.get(path, relentry=relentry)


# Note: We subclass from VirtualDirectory because VirtualFilesystem is currently only compatible with VirtualDirectory
# Subclass from VirtualDirectory so we get that compatibility for free, and override the rest to do our own thing
class ZipFilesystemEntry(VirtualDirectory):
    fs: ZipFilesystem
    entry: zipfile.ZipInfo

    def __init__(self, fs: ZipFilesystem, path: str, entry: zipfile.ZipInfo):
        super().__init__(fs, path)
        self.entry = entry

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)

        if self.is_symlink():
            return self._resolve().open()

        try:
            return BufferedStream(self.fs.zip.open(self.entry), size=self.entry.file_size)
        except Exception:
            raise FileNotFoundError(self.path)

    def iterdir(self) -> Iterator[str]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        entry = self._resolve()
        if isinstance(entry, ZipFilesystemEntry):
            yield from super(ZipFilesystemEntry, entry).iterdir()
        else:
            yield from entry.iterdir()

    def scandir(self) -> Iterator[FilesystemEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        entry = self._resolve()
        if isinstance(entry, ZipFilesystemEntry):
            yield from super(ZipFilesystemEntry, entry).scandir()
        else:
            yield from entry.scandir()

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            entry = self._resolve(follow_symlinks=follow_symlinks)
        except FilesystemError:
            return False

        if isinstance(entry, ZipFilesystemEntry):
            return entry.entry.is_dir()
        return isinstance(entry, VirtualDirectory)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            entry = self._resolve(follow_symlinks=follow_symlinks)
        except FilesystemError:
            return False

        if isinstance(entry, ZipFilesystemEntry):
            return not entry.entry.is_dir()
        return False

    def is_symlink(self) -> bool:
        return stat.S_ISLNK(self.entry.external_attr >> 16)

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError
        return self.fs.zip.open(self.entry).read().decode()

    def readlink_ext(self) -> FilesystemEntry:
        return FilesystemEntry.readlink_ext(self)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
        mode = self.entry.external_attr >> 16

        if self.entry.is_dir() and not stat.S_ISDIR(mode):
            mode = stat.S_IFDIR | mode
        elif not self.entry.is_dir() and not stat.S_ISREG(mode):
            mode = stat.S_IFREG | mode

        try:
            mtime = datetime(*self.entry.date_time, tzinfo=timezone.utc)
        except ValueError:
            mtime_tuple = (2107, 12, 31, 23, 59, 59) if self.entry.date_time[0] >= 2107 else (1980, 1, 1, 0, 0, 0)
            mtime = datetime(*mtime_tuple, tzinfo=timezone.utc)

        return fsutil.stat_result(
            [
                mode,
                self.entry.header_offset,
                id(self.fs),
                1,
                0,
                0,
                self.entry.file_size,
                0,
                mtime.timestamp(),
                0,
            ]
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Generic, NamedTuple, TypeVar


class ProcedureDescriptor(NamedTuple):
    program: int
    version: int
    procedure: int


GetPortProc = ProcedureDescriptor(100000, 2, 3)


class Bool(IntEnum):
    FALSE = 0
    TRUE = 1


class AcceptStat(IntEnum):
    SUCCESS = 0  # RPC executed successfully
    PROG_UNAVAIL = 1  # remote hasn't exported program
    PROG_MISMATCH = 2  # remote can't support version #
    PROC_UNAVAIL = 3  # program can't support procedure
    GARBAGE_ARGS = 4  # procedure can't decode params


class RejectStat(IntEnum):
    RPC_MISMATCH = 0
    AUTH_ERROR = 1


class AuthStat(IntEnum):
    AUTH_BADCRED = 1  # bad credentials (seal broken)
    AUTH_REJECTEDCRED = 2  # client must begin new session
    AUTH_BADVERF = 3  # bad verifier (seal broken)
    AUTH_REJECTEDVERF = 4  # verifier expired or replayed
    AUTH_TOOWEAK = 5  # rejected for security reasons


ProcedureParams = TypeVar("ProcedureParams")
ProcedureResults = TypeVar("ProcedureResults")
Credentials = TypeVar("Credentials")
Verifier = TypeVar("Verifier")


@dataclass
class AuthNull:
    pass


@dataclass
class AuthUnix:
    stamp: int  # Arbitrary id
    machinename: str
    uid: int
    gid: int
    gids: list[int]

    def __post_init__(self):
        if len(self.gids) > 16:
            raise ValueError("gids list cannot exceed 16 elements")

        if len(self.machinename) > 255:
            raise ValueError("machinename cannot exceed 255 characters")


@dataclass
class CallBody(Generic[ProcedureParams, Credentials, Verifier]):
    program: int
    version: int
    procedure: int
    cred: Credentials
    verf: Verifier
    params: ProcedureParams
    rpc_version: int = 2


@dataclass
class Mismatch:
    low: int
    high: int


@dataclass
class AcceptedReply(Generic[ProcedureResults, Verifier]):
    verf: Verifier
    stat: AcceptStat
    results: ProcedureResults | Mismatch | None


@dataclass
class RejectedReply:
    stat: RejectStat
    result: Mismatch | AuthStat


@dataclass
class Message(Generic[ProcedureParams, ProcedureResults, Credentials, Verifier]):
    xid: int
    body: CallBody[ProcedureParams, Credentials, Verifier] | AcceptedReply[ProcedureResults, Verifier] | RejectedReply


class Protocol(IntEnum):
    TCP = 6
    UDP = 17


@dataclass
class PortMapping:
    program: int
    version: int
    protocol: Protocol
    port: int = 0



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target import Target
from dissect.target.loader import open as loader_open
from dissect.target.loaders.overlay2 import Overlay2Loader

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], target_linux_docker: Target) -> None:
    """Test that we correctly use ``Overlay2Loader`` when opening a ``Target``."""
    for container in [
        "589135d12011921ac6ce69753569da5f206f4bc792a9133727ddae860997ee66",
        "f988f88e221d97930a665712cf16ab520f7e2c5af395660c145df93aebedf071",
    ]:
        path = target_linux_docker.fs.path("/var/lib/docker/image/overlay2/layerdb/mounts").joinpath(container)
        target = opener(path)
        assert isinstance(target._loader, Overlay2Loader)
        assert target.path == path


@pytest.mark.parametrize(
    ("path", "layers", "entries"),
    [
        ("589135d12011921ac6ce69753569da5f206f4bc792a9133727ddae860997ee66", 4, 18),
        ("f988f88e221d97930a665712cf16ab520f7e2c5af395660c145df93aebedf071", 9, 19),
    ],
)
def test_docker(target_linux_docker: Target, path: str, layers: int, entries: int) -> None:
    """Test if we correctly detect and map a Docker container."""
    container = target_linux_docker.fs.path("/var/lib/docker/image/overlay2/layerdb/mounts").joinpath(path)

    loader = loader_open(container)
    assert isinstance(loader, Overlay2Loader)

    t = Target()
    loader.map(t)

    assert len(t.filesystems) == 1

    container_fs = t.filesystems[0]
    assert len(container_fs.layers) == layers
    assert len(list(container_fs.path("/").iterdir())) == entries



from unittest.mock import patch

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.esxi._os import (
    _create_local_fs,
    _decrypt_crypto_util,
    esxi_hash,
)
from dissect.target.target import Target


def test__create_tar_fs_no_envelope(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    with (
        patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", False),
        patch("dissect.target.plugins.os.unix.esxi._os.tar") as mocked_tar,
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_crypto_util") as decrypt_func,
    ):
        target_linux._name = "local"
        _create_local_fs(target_linux, fs_unix.path("local.tgz.ve"), fs_unix.path("encryption.info"))

        decrypt_func.assert_called()
        mocked_tar.TarFilesystem.assert_called()


def test__create_tar_fs_envelope(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    with (
        patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", True),
        patch("dissect.target.plugins.os.unix.esxi._os.tar") as mocked_tar,
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_envelope") as decrypt_func,
    ):
        _create_local_fs(target_linux, fs_unix.path("local.tgz.ve"), fs_unix.path("encryption.info"))

        decrypt_func.assert_called()
        mocked_tar.TarFilesystem.assert_called()


def test__create_tar_fs_failed_envelope(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    with (
        patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", True),
        patch("dissect.target.plugins.os.unix.esxi._os.tar") as mocked_tar,
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_envelope", side_effect=[NotImplementedError]),
        patch("dissect.target.plugins.os.unix.esxi._os._decrypt_crypto_util") as decrypt_func,
    ):
        target_linux._name = "local"
        _create_local_fs(target_linux, fs_unix.path("local.tgz.ve"), fs_unix.path("encryption.info"))

        decrypt_func.assert_called()
        mocked_tar.TarFilesystem.assert_called()


def test__decrypt_crypto_not_local(target_linux: Target, fs_unix: VirtualFilesystem) -> None:
    target_linux._name = "not_local"
    with patch("dissect.target.plugins.os.unix.esxi._os.HAS_ENVELOPE", False):
        assert _create_local_fs(target_linux, fs_unix.path(""), fs_unix.path("")) is None


def test__decrypt_crypto_local(fs_unix: VirtualFilesystem) -> None:
    with patch("dissect.target.plugins.os.unix.esxi._os.subprocess.run") as mocked_run:
        mocked_run.return_value.stdout = b"data"

        assert _decrypt_crypto_util(fs_unix.path("data")).read() == b"data"


def test_hash_remainder() -> None:
    ip = b"192.168.1.109"
    volume = b"/home/roel/nfstest"
    h1 = esxi_hash(ip, 42)
    h2 = esxi_hash(volume, h1)
    assert h2 == 5364432747070711354


def test_hash_full() -> None:
    h1 = esxi_hash(b"Het implementeren van hashfuncties in Python is lastiger dan je zou denken,", 42)
    h2 = esxi_hash(b"met name door de ontbrekende ondersteuning voor unsigned integer arithmetic", h1)
    assert h2 == 2809036171121327430


def test_hash_empty_key() -> None:
    h = esxi_hash(b"", 666)
    assert h == 8664614747486377173



from __future__ import annotations

import io
from typing import TYPE_CHECKING

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

adtev_def = """
struct header {
    uint16  unk0;
    uint16  unk1;
    uint16  num_categories;
    uint16  unk2;
    uint16  footer_offset;
    uint16  unk3;
};
"""

c_adtev = cstruct().load(adtev_def)

POLICY_CATEGORIES = [
    "System",
    "Logon/Logoff",
    "Object Access",
    "Privilege Use",
    "Detailed Tracking",
    "Policy Change",
    "Account Management",
    "DS Access",
    "Account Logon",
]

POLICY_MAP = {
    "System": [
        "Security State Change",
        "Security System Extension",
        "System Integrity",
        "IPsec Driver",
        "Other System Events",
    ],
    "Logon/Logoff": [
        "Logon",
        "Logoff",
        "Account Lockout",
        "IPsec Main Mode",
        "Special Logon",
        "IPsec Quick Mode",
        "IPsec Extended Mode",
        "Other Logon/Logoff Events",
        "Network Policy Server",
        "User/Device Claims",
        "Group Membership",
    ],
    "Object Access": [
        "File System",
        "Registry",
        "Kernel Object",
        "SAM",
        "Other Object Access Events",
        "Certification Services",
        "Application Generated",
        "Handle Manipulation",
        "File Share",
        "Filtering Platform Packet Drop",
        "Filtering Platform Connection",
        "Detailed File Share",
        "Removable Storage",
        "Central Policy Staging",
    ],
    "Privilege Use": [
        "Sensitive Privilege Use",
        "Non Sensitive Privilege Use",
        "Other Privilege Use Events",
    ],
    "Detailed Tracking": [
        "Process Creation",
        "Process Termination",
        "DPAPI Activity",
        "RPC Events",
        "Plug and Play Events",
        "Token Right Adjusted Events",
    ],
    "Policy Change": [
        "Audit Policy Change",
        "Authentication Policy Change",
        "Authorization Policy Change",
        "MPSSVC Rule-Level Policy Change",
        "Filtering Platform Policy Change",
        "Other Policy Change Events",
    ],
    "Account Management": [
        "User Account Management",
        "Computer Account Management",
        "Security Group Management",
        "Distribution Group Management",
        "Application Group Management",
        "Other Account Management Events",
    ],
    "DS Access": [
        "Directory Service Access",
        "Directory Service Changes",
        "Directory Service Replication",
        "Detailed Directory Service Replication",
    ],
    "Account Logon": [
        "Credential Validation",
        "Kerberos Service Ticket Operations",
        "Other Account Logon Events",
        "Kerberos Authentication Service",
    ],
}

POLICY_VALUES = {
    0: "No auditing",
    1: "Success",
    2: "Failure",
    3: "Success/Failure",
}

AuditPolicyRecord = TargetRecordDescriptor(
    "windows/registry/auditpol",
    [
        ("string", "category"),
        ("string", "name"),
        ("string", "value"),
    ],
)


class AuditpolPlugin(Plugin):
    """Plugin that parses audit policy settings from the registry."""

    KEY = "HKLM\\SECURITY\\Policy\\PolAdtEv"

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEY))) > 0:
            raise UnsupportedPluginError(f"Registry key {self.KEY} not found")

    @export(record=AuditPolicyRecord)
    def auditpol(self) -> Iterator[AuditPolicyRecord]:
        """Return audit policy settings from the registry.

        For Windows, the audit policy settings are stored in the HKEY_LOCAL_MACHINE\\Security\\Policy\\PolAdtEv registry
        key. It shows for each possible audit event if it is logged.

        References:
            - https://countuponsecurity.com/tag/poladtev/
        """
        for regf in self.target.registry.keys(self.KEY):
            for entry in regf.values():
                if not len(entry.value):
                    continue

                buf = io.BytesIO(entry.value)
                header = c_adtev.header(buf)
                data_offset = buf.tell()

                buf.seek(header.footer_offset)
                category_counts = c_adtev.uint16[header.num_categories](buf)

                buf.seek(data_offset)
                for i in range(header.num_categories):
                    num_entries = category_counts[i]
                    category_name = POLICY_CATEGORIES[i]
                    labels = POLICY_MAP[category_name]
                    values = c_adtev.uint16[category_counts[i]](buf)

                    for j in range(num_entries):
                        try:
                            label = labels[j]
                        except IndexError:
                            label = "Unknown"

                        yield AuditPolicyRecord(
                            category=category_name,
                            name=label,
                            value=POLICY_VALUES[values[j]],
                            _target=self.target,
                        )



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.hypervisor import vdi

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class VdiContainer(Container):
    """VirtualBox hard disks."""

    __type__ = "vdi"

    def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
        f = fh
        if not hasattr(fh, "read"):
            f = fh.open("rb")
        self.vdi = vdi.VDI(f)

        super().__init__(fh, self.vdi.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return fh.read(68)[-4:] == b"\x7f\x10\xda\xbe"

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return path.suffix.lower() == ".vdi"

    def read(self, length: int) -> bytes:
        return self.vdi.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.vdi.seek(offset, whence)

    def tell(self) -> int:
        return self.vdi.tell()

    def close(self) -> None:
        pass



from __future__ import annotations

from typing import Union

from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import NamespacePlugin

COMMON_FIELDS = [
    ("datetime", "ts"),
    ("string", "client"),
    ("string", "account"),
    ("string", "sender"),
    ("string", "recipient"),
]

GENERIC_USER_FIELDS = [
    ("datetime", "ts_mtime"),
    ("string", "client"),
    ("string", "account"),
]

GENERIC_ATTACHMENT_FIELDS = [
    *COMMON_FIELDS,
    ("path", "attachment"),
    ("string", "description"),
]

GENERIC_MESSAGE_FIELDS = [
    *COMMON_FIELDS,
    ("string", "message"),
]

ChatUserRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "chat/user",
    GENERIC_USER_FIELDS,
)

ChatMessageRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "chat/message",
    GENERIC_MESSAGE_FIELDS,
)

ChatAttachmentRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "chat/attachment",
    GENERIC_ATTACHMENT_FIELDS,
)

ChatRecord = Union[ChatUserRecord, ChatMessageRecord, ChatAttachmentRecord]


class ChatPlugin(NamespacePlugin):
    __namespace__ = "chat"
